<?php

/**
 *
 * mpath-tools - failover and load balancing between multiple ISP made easy
 * 
 * Copyright (C) 2008-2010 Vincent Negrier aka. sIX <six at aegis-corp.org>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 *
 * @package mpath-tools
 * @subpackage control
 */

namespace MPath_Tools\Control;

require "mpath-tools/inc/nanoserv/nanoserv.php";
require "mpath-tools/inc/control.php";
require "mpath-tools/inc/control_client.php";

use MPath_Tools\Control\Client_Connection;
use MPath_Tools\Control\Client_Options;
use MPath_Tools\Control\Packet;

use \Nanoserv\Core as Nanoserv;
use \Exception;

class Client_Exception extends Exception {

}

class Client_Events {

	public $callbacks = array(	"on_Hello"			=> false,
								"on_Log_Message"	=> false,
								"on_Disconnect"		=> false);

	public function Register($ev_name, $ev_handler) {

		if (!isset($this->callbacks[$ev_name])) {

			throw new Client_Exception("cannot register unknown event '{$ev_name}'");

		} else if (!is_callable($ev_handler)) {

			throw new Client_Exception("invalid handler for event '{$ev_name}'");

		}

		$this->callbacks[$ev_name] = $ev_handler;
	
	}

	public function Unregister($ev_name) {

		if (!isset($this->callbacks[$ev_name])) {

			throw new Client_Exception("cannot unregister unknown event '{$ev_name}'");

		}

		$this->callbacks[$ev_name] = false;
	
	}
	
	public function Dispatch($ev_name, $params = NULL) {

		if (!isset($this->callbacks[$ev_name])) {

			throw new Client_Exception("cannot dispatch unknown event '{$ev_name}' (internal error)");

		} else if ($this->callbacks[$ev_name] !== false) {
		
			call_user_func_array($this->callbacks[$ev_name], $params);

		}

	}

	public function __set($name, $value) {

		$this->Register($name, $value);
	
	}

	public function __unset($name) {

		$this->Unregister($name);
	
	}
	
	public function __call($name, $params) {

		$this->Dispatch($name, $params);

	}

}

class Client {

	const DEFAULT_HOSTNAME = "localhost";
	const DEFAULT_PORT = 7842;
	const DEFAULT_AGENT = "mpath-client";
	
	public $hostname = self::DEFAULT_HOSTNAME;
	public $port = self::DEFAULT_PORT;
	public $options;
	public $events;

	public $server;

	public $ready = false;

	private $ctl;
	private $requeue = array();
	
	public function __construct($hostname = NULL, $port = NULL) {

		if (isset($hostname)) {

			$this->hostname = $hostname;
		
		}

		if (isset($port)) {

			if (!is_numeric($port)) {

				throw new Client_Exception("invalid port number '{$port}'");

			}

			$this->port = $port;

		}

		$this->options = new Client_Options(self::DEFAULT_AGENT);
		$this->events = new Client_Events();
	
	}

	public function Poll($time = 1, $user_streams = array()) {

		if (isset($this->requeue)) {

			$ret = $this->requeue;
			unset($this->requeue);

		} else {
		
			$ret = array();

		}
		
		Nanoserv::Run($time, $user_streams);

		while ($pkt = $this->ctl->Get_Packet()) {

			switch ($pkt->type) {

				case Packet::TYPE_HELLO:
				$this->server = $pkt->message["server"];
				$this->ready = true;
				$this->events->on_Hello();
				break;

				case Packet::TYPE_LOG:
				$this->events->on_Log_Message($pkt->message["level"], $pkt->message["message"]);
				break;
				
				default:
				$ret[] = $pkt;
			
			}
		
		}

		if (!$this->ctl->socket->connected) {

			$this->ready = false;
			$this->events->on_Disconnect();
		
		}

		return $ret;
		
	}

	public function Connect($timeout = 5) {

		$this->ctl = Nanoserv::New_Connection("tcp://{$this->hostname}:{$this->port}", __NAMESPACE__ . "\\Client_Connection", $this->options);
		$this->ctl->Connect($timeout);

		while (!$this->ready) {

			$this->Poll();

		}

	}

	public function Disconnect() {

		if (isset($this->ctl)) {
		
			Nanoserv::Flush_Write_Buffers();
			
			$this->ctl->Disconnect();

		}

		$this->ready = false;
	
	}

	public function __destruct() {

		$this->Disconnect();

	}

	private function Do_Request($type, $message = NULL, $timeout = 5) {

		$requeue = array();
		$max_t = time() + $timeout;
		
		$this->ctl->Send($type, $message);

		while ((time() < $max_t) && (!isset($ret))) {

			foreach ($this->Poll() as $tmp) {

				if ($tmp->type === $type) {

					$ret = $tmp->message;
				
				} else {

					$requeue[] = $tmp;

				}
		
			}
		
		}

		if (!isset($ret)) {
		
			throw new Client_Exception("server timeout (request type {$type})");
			
		}
		
		$this->requeue = $requeue;

		if (is_array($ret) && isset($ret["error"])) {

			throw new Client_Exception("server error: {$ret['error']}");

		}
		
		return $ret;

	}

	// API Starts here
	
	public function Stop_Daemon($message = NULL) {

		$this->ctl->Send(Packet::TYPE_STOP, $message);
	
	}

	public function Refresh_Rules() {

		$this->ctl->Send(Packet::TYPE_REFRESH);
	
	}

	public function Commit_Rules() {

		return $this->Do_Request(Packet::TYPE_COMMIT);
	
	}

	public function List_Gateways($search = NULL) {

		return $this->Do_Request(Packet::TYPE_GATEWAY_LIST, $search);
	
	}

	public function Get_Gateways(array $names) {

		return $this->Do_Request(Packet::TYPE_GATEWAY_GET, $names);
	
	}

	public function Set_Gateway($gwname, array $updates) {

		return $this->Do_Request(Packet::TYPE_GATEWAY_SET, array("gateway" => $gwname, "updates" => $updates));

	}

	public function Get_Rules(array $rule = NULL) {

		return $this->Do_Request(Packet::TYPE_RULE_GET, $rule);

	}

	public function Add_Rule(array $rule) {

		return $this->Do_Request(Packet::TYPE_RULE_ADD, $rule);

	}

	public function Insert_Rule(array $rule, $position = 1) {

		return $this->Do_Request(Packet::TYPE_RULE_INSERT, array("rule" => $rule, "position" => $position));
	
	}
	
	public function Del_Rule(array $rule) {

		return $this->Do_Request(Packet::TYPE_RULE_DEL, $rule);
	
	}

	public function Set_Rule(array $rule) {

		return $this->Do_Request(Packet::TYPE_RULE_SET, $rule);
	
	}

	public function Get_Log($filename) {

		return $this->Do_Request(Packet::TYPE_LOG_GET, $filename);
	
	}
	
	public function Set_Log(array $log) {

		return $this->Do_Request(Packet::TYPE_LOG_SET, $log);
	
	}

	public function Reload() {

		return $this->Do_Request(Packet::TYPE_RELOAD);
	
	}

	public function List_Controls() {

		return $this->Do_Request(Packet::TYPE_CONTROL_LIST);
	
	}

	public function Get_Status() {

		return $this->Do_Request(Packet::TYPE_STATUS);
	
	}

}

?>
