<?php

/**
 *
 * mpath-tools - failover and load balancing between multiple ISP made easy
 * 
 * Copyright (C) 2008-2010 Vincent Negrier aka. sIX <six at aegis-corp.org>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 *
 * @package mpath-tools
 * @subpackage config
 */

namespace MPath_Tools\Config;

require "mpath-tools/inc/tokenizer.php";

use \Exception;
use \Tokenizer\Exception as Tokenizer_Exception;
use \ArrayAccess, \IteratorAggregate, \ArrayIterator;

class Core {

	const DEFAULT_PATH = "/etc/mpath-tools";

	static public $gateways = array();
	
	static private $root;
	static private $cache = array();

	static private function Parse_Elements(Tokenizer $t, $conffile = NULL) {

		$tok = $t->Get_Token();

		if (!$tok->Is_Key()) {

			throw new Exception("unexpected {$tok->token} at line {$tok->line}: '{$tok->string}' (not a configuration directive)");
		
		}

		$name = $tok->string;
		
		if ($tok->Is_Value_Key()) {

			$el = self::Parse_Value($name, $t);

		} else if ($tok->Is_Block_Key()) {

			$el = self::Parse_Block($name, $t, $conffile);

		} else {

			throw new Exception("unexpected {$tok->token} at line {$tok->line}: '{$tok->string}' (internal error)");
		
		}
	
		if ($el->name === "include") {

			if (!isset($conffile)) {

				throw new Exception("cannot include when not parsing from file");

			}
			
			$inc = array();
			
			foreach ($el->Get_Values() as $inc_fn) {

				try {
				
					foreach (self::Parse_File(dirname($conffile) . DIRECTORY_SEPARATOR . $inc_fn) as $inc_el) {

						$inc[] = $inc_el;

					}

				} catch (Exception $e) {

					throw new Exception("include({$inc_fn}): {$e->getMessage()}");

				}

			}
		
			return $inc;
		
		} else {
		
			return array($el);

		}
	
	}
	
	static private function Parse_Block($name, Tokenizer $t, $conffile) {

		$values = self::_parse_value($t, "T_BLOCK_START");

		$elements = array();
		
		$tok = $t->Get_Token(Tokenizer::PEEK);
		
		while ($tok && ($tok->token !== "T_BLOCK_END")) {

			foreach (self::Parse_Elements($t, $conffile) as $el) {
			
				$elements[] = $el;

			}
		
			$tok = $t->Get_Token(Tokenizer::PEEK);
		
		}

		if ($tok === false) {

			throw new Exception("unexpected end of file");
		
		} else {

			$t->Get_Token();

		}

		return new Block($name, $values, $elements);
	
	}

	static private function _parse_value(Tokenizer $t, $delim) {

		$values = array();
		
		$tok = $t->Get_Token();

		while ($tok && ($tok->token !== $delim)) {

			if (!$tok->Is_Value()) {

				throw new Exception("unexpected {$tok->token} at line {$tok->line}: '{$tok->string}' (may be a missing ';')");
			
			}

			$values[] = $tok->Get_Value();
			
			$tok = $t->Get_Token();
		
		}

		if ($tok === false) {

			throw new Exception("unexpected end of file");
		
		}
		
		return $values;
	
	}
	
	static private function Parse_Value($name, Tokenizer $t) {

		$values = self::_parse_value($t, "T_SEMICOLON");
	
		return new Element($name, $values);
	
	}
	
	static public function Parse_String($str, $conffile = NULL) {

		$t = Tokenizer::Create($str);

		$conf = array();

		while ($t->Get_Token(Tokenizer::PEEK)) {

			foreach(self::Parse_Elements($t, $conffile) as $el) {

				$conf[] = $el;

			}

		}

		if (isset($conffile)) {
		
			return $conf;

		} else {

			return new Block("root", array(), $conf);

		}
	
	}
	
	static private function Parse_File($rconffile) {
		
		$conffile = realpath($rconffile);
		
		if (!isset(self::$cache[$conffile])) {
		
			if (!file_exists($conffile)) {

				throw new Exception("configuration file '{$rconffile}' not found");

			}

			if (!is_readable($conffile) || !is_file($conffile)) {

				throw new Exception("cannot read configuration file '{$conffile}'");

			}
			
			try {

				$conf = self::Parse_String(file_get_contents($conffile), $conffile);

			} catch (Tokenizer_Exception $e) {

				throw new Tokenizer_Exception("{$e->getMessage()} in '{$conffile}'");

			}
			
			self::$cache[$conffile] = $conf;

		}
		
		return self::$cache[$conffile];
	
	}
		
	static public function Parse($fn, $clear_cache = true) {

		if ($clear_cache) {
			
			self::Clear_Cache();

		}
		
		$conf = self::Parse_File($fn);
		
		$root = new Block("root", array(), $conf);

		if ($gateways = $root->Get_Elements_By_Name("gateway")) {
		
			self::$gateways = $gateways;

		}
		
		return $root;
	
	}

	static public function Query() {

		return self::$root;

	}

	static public function Get_Gateways() {

		return self::$gateways;

	}

	static public function Clear_Cache() {

		self::$cache = array();

	}

}

class Element implements ArrayAccess, IteratorAggregate {

	public $name;
	public $value = array();

	public function __construct($name, array $value) {

		$this->name = $name;
		$this->value = $value;

	}

	public function __toString() {

		return $this->Get_Value();
	
	}
	
	public function offsetExists($i) {

		return isset($this->value[$i]);

	}

	public function offsetGet($i) {

		return $this->Get_Value($i);

	}

	public function offsetSet($i, $value) {

		throw new Exception("cannot set configuration element '{$i}'");
	
	}

	public function offsetUnset($i) {

		throw new Exception("cannot unset configuration element '{$i}'");

	}
	
	public function getIterator() {

		return new ArrayIterator($this->value);
	
	}
	
	public function Get_Name() {

		return $this->name;
	
	}
	
	public function Get_Value($i = 0) {

		return $this->value[$i];

	}

	public function Get_Values() {

		return $this->value;

	}

}

class Block extends Element {

	public $elements = array();
	
	public function __construct($name, array $value, array $elements) {

		parent::__construct($name, $value);

		$this->elements = $elements;
	
	}

	public function __get($name) {

		return $this->Get_Element_By_Name($name);

	}

	public function __call($name, $params) {

		list($value) = $params;
		
		foreach ($this->Get_Elements_By_Name($name) as $el) {

			if (in_array($value, $el->Get_Values())) {

				return $el;

			}
		
		}

		return false;
	
	}

	public function Get_Element_By_Name($name) {

		$name = strtolower($name);
		
		foreach ($this->elements as $el) {

			if (strtolower($el->name) === $name) {

				return $el;

			}

		}

		return false;
	
	}

	public function Get_Elements_By_Name($name) {

		$ret = array();
		
		$name = strtolower($name);
		
		foreach ($this->elements as $el) {

			if (strtolower($el->name) === $name) {

				$ret[] = $el;

			}

		}

		return $ret;
	
	}

	public function getIterator() {

		return new ArrayIterator($this->elements);
	
	}

}

class Token extends \Tokenizer\Token {

	public function Is_String() {

		return in_array($this->token, array("T_STRING_D", 
											"T_STRING_S", 
											"T_STRING_B",
											"T_STRING_U"));
	
	}

	public function Is_Int() {

		return in_array($this->token, array("T_INTEGER"));
	
	}
	
	public function Is_Float() {

		return in_array($this->token, array("T_FLOAT"));

	}
	
	public function Is_Value_Key() {

		return in_array($this->token, array("T_IP_ADDR",
											"T_NETWORK",
											"T_NEXTHOP",
											"T_INTERFACE",
											"T_TEST",
											"T_UP",
											"T_DOWN",
											"T_INCLUDE",
											"T_STATE",
											"T_ACTIVE",
											"T_DISENGAGED",
											"T_DISABLED",
											"T_LOCAL",
											"T_RESET",
											"T_ALERT",
											"T_GATEWAYS",
											"T_INTERVAL",
											"T_SIZE",
											"T_PATTERN",
											"T_TIMEOUT",
											"T_COLOR",
											"T_LEVEL",
											"T_PIDFILE",
											"T_INTERVAL",
											"T_LATENCY",
											"T_DISENGAGE",
											"T_DISABLE",
											"T_ACTIVATE",
											"T_LISTEN",
											"T_WEIGHT",
											"T_VIA",
											"T_POLICY",
											"T_EXPIRES",
											"T_STARTUP",
											"T_SHUTDOWN",
											"T_TABLE",
											"T_NO_SOURCE_RULE"));
	
	}

	public function Is_Block_Key() {

		return in_array($this->token, array("T_GATEWAY",
											"T_DEFAULT",
											"T_SOURCE",
											"T_DESTINATION",
											"T_FWMARK",
											"T_TOS",
											"T_EVENTS",
											"T_COLORS",
											"T_DEFAULTS",
											"T_LOG",
											"T_PROBE",
											"T_CONTROL",
											"T_RULES"));
	
	}

	public function Is_Key() {

		return $this->Is_Value_Key() || $this->Is_Block_Key();
	
	}
	
	public function Is_Value() {

		return $this->Is_String() || $this->Is_Int() || $this->Is_Float();
	
	}

	public function Get_Value() {

		if (!$this->Is_Value()) {

			throw new Exception("cannot get value of a non value token");
		
		} else if ($this->Is_String()) {

			switch ($this->token) {

				case "T_STRING_U":
				case "T_STRING_S":		return (string)$this->data[0];
				case "T_STRING_D":		return stripcslashes($this->data[0]);
				case "T_STRING_B":		return `$this->data[0]`;

				default:				throw new Exception("unknown string type {$this->token} (internal error)");

			}

		} else if ($this->Is_Int()) {

			return (int)$this->data[0];

		} else if ($this->Is_Float()) {
			
			return (float)$this->data[0];
		
		} else {

			throw new Exception("unknown value type");
		
		}
	
	}

}

class Tokenizer extends \Tokenizer\Core {

	const T_COMMENT = "/#.*/";
	const T_WHITESPACE = "/[ \t\r\n]+/";

	const T_STRING_U = "/([^ \t\r\n;{}\"'`]+)/";
	const T_STRING_D = "/\"(.*[^\\\])\"/U";
	const T_STRING_S = "/'(.*[^\\\])'/U";
	const T_STRING_B = "/`(.*[^\\\])`/U";

	const T_INTEGER = "/(-?[0-9]+)/";
	const T_FLOAT = "/(-?[0-9]+\.[0-9]+)/";

	const T_BLOCK_START = "/\{/";
	const T_BLOCK_END = "/\}/";

	const T_SEMICOLON = "/;/";
	
	const T_INCLUDE = "/include/i";
	
	const T_GATEWAY = "/gateway/i";
	const T_IP_ADDR = "/ip_addr/i";
	const T_NETWORK = "/network/i";
	const T_NEXTHOP = "/nexthop/i";
	const T_INTERFACE = "/interface/i";
	const T_RULES = "/rules/i";
	const T_TEST = "/test/i";
	const T_STATE = "/state/i";
	const T_ACTIVE = "/active/i";
	const T_DISENGAGED = "/disengaged/i";
	const T_DISABLED = "/disabled/i";
	const T_EVENTS = "/events/i";
	const T_SOURCE = "/source/i";
	const T_DESTINATION = "/destination/i";
	const T_FWMARK = "/fwmark/i";
	const T_TOS = "/tos/i";
	const T_UP = "/up/i";
	const T_DOWN = "/down/i";
	const T_LOCAL = "/local/i";
	const T_COLORS = "/colors/i";
	const T_RESET = "/reset/i";
	const T_ALERT = "/alert/i";
	const T_GATEWAYS = "/gateways/i";
	const T_DEFAULTS = "/defaults/i";
	const T_INTERVAL = "/interval/i";
	const T_SIZE = "/size/i";
	const T_PATTERN = "/pattern/i";
	const T_TIMEOUT = "/timeout/i";
	const T_COLOR = "/color/i";
	const T_LOG = "/log/i";
	const T_LEVEL = "/level/i";
	const T_PIDFILE = "/pidfile/i";
	const T_PROBE = "/probe/i";
	const T_LATENCY = "/latency/i";
	const T_DISENGAGE = "/disengage/i";
	const T_DISABLE = "/disable/i";
	const T_ACTIVATE = "/activate/i";
	const T_CONTROL = "/control/i";
	const T_LISTEN = "/listen/i";
	const T_DEFAULT = "/default/i";
	const T_VIA = "/via/i";
	const T_POLICY = "/policy/i";
	const T_EXPIRES = "/expires/i";
	const T_WEIGHT = "/weight/i";
	const T_STARTUP = "/startup/i";
	const T_SHUTDOWN = "/shutdown/i";
	const T_TABLE = "/table/i";
	const T_NO_SOURCE_RULE = "/no_source_rule/i";

	const T_NETWORK_ALIASES = "/network|net/i";
	const T_INTERFACE_ALIASES = "/interface|if|dev/i";
	const T_SOURCE_ALIASES = "/source|src|from/i";
	const T_DESTINATION_ALIASES = "/destination|dest|dst|to/i";
	const T_VIA_ALIASES = "/via|gw/i";

	
	public $tokens = array(	"T_COMMENT"			=> self::T_COMMENT,
							"T_GATEWAY"			=> self::T_GATEWAY,
							"T_IP_ADDR"			=> self::T_IP_ADDR,
							"T_NETWORK"			=> self::T_NETWORK,
							"T_NEXTHOP"			=> self::T_NEXTHOP,
							"T_INTERFACE"		=> self::T_INTERFACE,
							"T_RULES"			=> self::T_RULES,
							"T_TEST"			=> self::T_TEST,
							"T_EVENTS"			=> self::T_EVENTS,
							"T_SOURCE"			=> self::T_SOURCE,
							"T_DESTINATION"		=> self::T_DESTINATION,
							"T_FWMARK"			=> self::T_FWMARK,
							"T_TOS"				=> self::T_TOS,
							"T_UP"				=> self::T_UP,
							"T_DOWN"			=> self::T_DOWN,
							"T_STATE"			=> self::T_STATE,
							"T_ACTIVE"			=> self::T_ACTIVE,
							"T_DISENGAGED"		=> self::T_DISENGAGED,
							"T_DISABLED"		=> self::T_DISABLED,
							"T_LOCAL"			=> self::T_LOCAL,
							"T_COLORS"			=> self::T_COLORS,
							"T_RESET"			=> self::T_RESET,
							"T_ALERT"			=> self::T_ALERT,
							"T_GATEWAYS"		=> self::T_GATEWAYS,
							"T_DEFAULTS"		=> self::T_DEFAULTS,
							"T_INTERVAL"		=> self::T_INTERVAL,
							"T_SIZE"			=> self::T_SIZE,
							"T_PATTERN"			=> self::T_PATTERN,
							"T_TIMEOUT"			=> self::T_TIMEOUT,
							"T_COLOR"			=> self::T_COLOR,
							"T_LOG"				=> self::T_LOG,
							"T_LEVEL"			=> self::T_LEVEL,
							"T_PIDFILE"			=> self::T_PIDFILE,
							"T_PROBE"			=> self::T_PROBE,
							"T_LATENCY"			=> self::T_LATENCY,
							"T_DISENGAGE"		=> self::T_DISENGAGE,
							"T_DISABLE"			=> self::T_DISABLE,
							"T_ACTIVATE"		=> self::T_ACTIVATE,
							"T_CONTROL"			=> self::T_CONTROL,
							"T_LISTEN"			=> self::T_LISTEN,
							"T_DEFAULT"			=> self::T_DEFAULT,
							"T_VIA"				=> self::T_VIA,
							"T_POLICY"			=> self::T_POLICY,
							"T_EXPIRES"			=> self::T_EXPIRES,
							"T_WEIGHT"			=> self::T_WEIGHT,
							"T_STARTUP"			=> self::T_STARTUP,
							"T_SHUTDOWN"		=> self::T_SHUTDOWN,
							"T_TABLE"			=> self::T_TABLE,
							"T_NO_SOURCE_RULE"  => self::T_NO_SOURCE_RULE,
							"T_INCLUDE"			=> self::T_INCLUDE,
							"T_WHITESPACE"		=> self::T_WHITESPACE,
							"T_BLOCK_START"		=> self::T_BLOCK_START,
							"T_BLOCK_END"		=> self::T_BLOCK_END,
							"T_SEMICOLON"		=> self::T_SEMICOLON,
							"T_INTEGER"			=> self::T_INTEGER,
							"T_FLOAT"			=> self::T_FLOAT,
							"T_STRING_D"		=> self::T_STRING_D,
							"T_STRING_S"		=> self::T_STRING_S,
							"T_STRING_B"		=> self::T_STRING_B);

	public function __construct($txt = NULL, $strict = true) {

		if (!$strict) {

			// Allow unquoted strings in non strict mode
			$this->Register_Token("T_STRING_U", self::T_STRING_U);

			// Allow shortcuts and aliases
			$this->Register_Token("T_NETWORK", self::T_NETWORK_ALIASES);
			$this->Register_Token("T_INTERFACE", self::T_INTERFACE_ALIASES);
			$this->Register_Token("T_SOURCE", self::T_SOURCE_ALIASES);
			$this->Register_Token("T_DESTINATION", self::T_DESTINATION_ALIASES);
			$this->Register_Token("T_VIA", self::T_VIA_ALIASES);

		}

		parent::__construct($txt, __NAMESPACE__ . "\\Token");

	}

	public function Get_Token($flags = 0) {

		if ($flags & Tokenizer::PEEK) {

			$p = $this->Get_Pointer();

		}
		
		$bypass_tokens = array(	"T_COMMENT",
								"T_WHITESPACE");
		
		do {

			$tok = parent::Get_Token();
	
		} while (in_array($tok->token, $bypass_tokens));

		if ($flags & Tokenizer::PEEK) {
		
			$this->Set_Pointer($p);

		}

		return $tok;
	
	}
	
	static public function Create($txt) {
		
		return new self($txt);
	
	}

}

?>
