<?php

/**
 *
 * mpath-tools - failover and load balancing between multiple ISP made easy
 * 
 * Copyright (C) 2008-2010 Vincent Negrier aka. sIX <six at aegis-corp.org>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 *
 * @package mpath-tools
 * @subpackage core
 */

namespace MPath_Tools;

const VERSION = "1.0.0";

use \Exception;

class Address {

	public $family;
	public $value;

	public function __construct($value) {

		$this->family = self::Get_Family($value);
		$this->value = $value;
	
	}

	public function Match_Network($network) {

		$net_addr = strtok($network, "/");
		$net_cidr = (int)strtok("");
		
		switch ($this->family) {

			case AF_INET:

			if (!self::Is_IPv4($net_addr)) return false;
			
			$addr_i = ip2long($this->value);
			$net_i1 = ip2long($net_addr);
			$net_i2 = $net_i1 + pow(2, 32 - $net_cidr) - 1;

			return (($net_i1 <= $addr_i) && ($addr_i <= $net_i2));
			
			
			case AF_INET6:

			if (!self::Is_IPv6($net_addr)) return false;
				
			$addr_s = inet_pton($this->value);
			$net_s = inet_pton($net_addr);

			$p8 = $net_cidr >> 3;

			$a_tmp = ord(substr($addr_s, $p8 - 1, 1));
			$n_tmp = ord(substr($net_s, $p8 -1, 1));

			$a2_tmp = substr($addr_s, 0, $p8 - 1);
			$n2_tmp = substr($net_s, 0, $p8 - 1);

			if ($a2_tmp !== $n2_tmp) return false;

			$mask = pow(2, 8 - ($net_cidr % 8)) - 1;
			
			return ($a_tmp & $mask) === ($n_tmp & $mask);

		}
	
	}
	
	public function Is_Local() {

		return ($this->value === "127.0.0.1") || ($this->value === "::1");
	
	}
	
	static public function Get($addr) {

		return new self($addr);
	
	}
	
	static public function Is_IPv4($addr) {

		return ip2long($addr) !== false;
	
	}

	static public function Is_IPv6($addr) {

		return preg_match('/^[0-9a-fA-F:]+$/', $addr) && @inet_pton($addr);
	
	}

	static public function Is_IP($addr) {

		return self::Is_IPv4($addr) || self::Is_IPv6($addr);

	}

	static public function Get_Family($addr) {

		if (self::Is_IPv4($addr)) {

			return AF_INET;

		} else if (self::Is_IPv6($addr)) {

			return AF_INET6;

		} else {

			throw new Exception("invalid address '{$addr}'");

		}
	
	}

}

class Gateway {

	const DEFAULT_DISENGAGE = 10;
	const DEFAULT_DISABLE = 60;
	const DEFAULT_ACTIVATE = 5;
	const DEFAULT_LATENCY = 500;
	
	const STATE_DISABLED = 0;
	const STATE_DISENGAGED = 1;
	const STATE_ACTIVE = 2;
	const STATE_AUTO = 255;
	
	const STATUS_UNKNOWN = 0;
	const STATUS_OK = 1;
	const STATUS_TIMEOUT = 2;
	const STATUS_ERROR = 3;
	const STATUS_HIGH_LATENCY = 4;
	
	const EVENT_ACTIVATE = 1;
	const EVENT_DISENGAGE = 2;
	const EVENT_DISABLE = 3;
	
	const PROTO_IPV4 = 1;
	const PROTO_IPV6 = 2;
	
	public $name;
	public $idx;

	public $ip_addr;
	public $network;
	public $nexthop;
	public $interface;

	public $weight = 1;

	public $proto = self::PROTO_IPV4;
	
	public $disengage = self::DEFAULT_DISENGAGE;
	public $disable = self::DEFAULT_DISABLE;
	public $activate = self::DEFAULT_ACTIVATE;
	public $max_latency = self::DEFAULT_LATENCY;

	public $admin_state = self::STATE_AUTO;
	public $state = self::STATE_ACTIVE;

	public $status = self::STATUS_UNKNOWN;
	public $status_change_t = 0;

	public $source_table;
	public $realm;
	
	public $latency;
	
	public $conf;
	public $route_source_network = true;
	
	public $events = array();
	
	static public $all = array();

	public function __construct($name, $ip_addr = NULL) {

		$this->name = $name;

		if (isset($ip_addr)) {

			$this->ip_addr = $ip_addr;

		}

	}

	public function Get_State() {

		if ($this->admin_state === self::STATE_AUTO) {

			return $this->state;

		} else {

			return $this->admin_state;

		}
	
	}
	
	public function Get_State_Str() {

		switch ($this->Get_State()) {

			case self::STATE_DISABLED:		return "disabled";
			case self::STATE_DISENGAGED:	return "disengaged";
			case self::STATE_ACTIVE:		return "active";

			default:						throw new Exception("unknown state '{$this->Get_State()}' (internal error)");

		}
	
	}

	public function Get_Status_Str() {

		if ($this->admin_state === self::STATE_AUTO) {
		
			switch ($this->status) {

				case self::STATUS_UNKNOWN:		return "unknown";
				case self::STATUS_OK:			return "ok (" . number_format($this->latency * 1000) . " ms)";
				case self::STATUS_TIMEOUT:		return "timeout";
				case self::STATUS_ERROR:		return "error";
				case self::STATUS_HIGH_LATENCY:	return "high-latency (" . number_format($this->latency * 1000) . " ms)";

			}

			throw new Exception("unknown status '{$this->status}' (internal error)");

		} else {

			return "admin";

		}
	
	}

	public function Prepare() {

		// Table for source based routing
		
		if ($tmp = Iproute2_Table_Info::Get_By_Name($this->name)) {

			$this->source_table = $tmp;

		} else {

			$this->source_table = Iproute2_Table_Info::Alloc($this->name, true);

		}

		// Routing realm

		if ($tmp = Iproute2_Realm::Get_By_Name($this->name)) {

			$this->realm = $tmp;

		}
	
	}

	public function Set_Status($status, $latency = NULL) {

		if (isset($latency)) {

			$this->latency = $latency;

		}
		
		$status = $this->Get_Status_From_Probe($status);
		
		if ($this->status !== $status) {

			$this->status = $status;
			$this->status_change_t = time();
		
		}

	}

	public function Set_Latency($latency) {

		$this->latency = $latency;

		if (($tmp = $this->Get_Status_From_Probe($this->status)) !== $this->status) {

			$this->Set_Status($tmp);

		}

	}

	public function Get_Status_From_Probe($status, $latency = NULL) {

		if (!isset($latency)) {

			$latency = $this->latency;

		}
		
		if (($status === self::STATUS_OK) && ($latency > $this->max_latency)) {

			$status = self::STATUS_HIGH_LATENCY;
		
		}

		return $status;
	
	}

	public function Set_Admin_State_Str($val) {

		switch (strtolower($val)) {

			case "active":		$state = self::STATE_ACTIVE;		break;
			case "disengaged":	$state = self::STATE_DISENGAGED;	break;
			case "disabled":	$state = self::STATE_DISABLED;		break;
			case "auto":		$state = self::STATE_AUTO;			break;

			default:			throw new Exception("invalid state '{$val}'");
		
		}

		$this->admin_state = $state;
	
	}
	
	public function Update_State($t = NULL) {

		if ($this->status === self::STATUS_UNKNOWN) {

			return false;

		}
		
		if (!isset($t)) {
			
			$t = time();

		}
		
		$newstate = false;
		
		if ($this->admin_state === self::STATE_AUTO) {
		
			if (($this->state !== self::STATE_DISABLED) && ($this->status !== self::STATUS_OK) && ($this->status !== self::STATUS_HIGH_LATENCY) && ($t >= ($this->status_change_t + $this->disable))) {

				$newstate = self::STATE_DISABLED;

			} else if (($this->state === self::STATE_ACTIVE) && ($this->status !== self::STATUS_OK) && ($t >= ($this->status_change_t + $this->disengage))) {

				$newstate = self::STATE_DISENGAGED;
			
			} else if (($this->state !== self::STATE_ACTIVE) && ($this->status === self::STATUS_OK) && ($t >= ($this->status_change_t + $this->activate))) {

				$newstate = self::STATE_ACTIVE;

			}

		} else if ($this->state !== $this->admin_state) {

			$newstate = $this->admin_state;

		}
		
		if ($newstate !== false) {
			
			$this->state = $newstate;

			return true;

		}

		return false;

	}

	public function Get_Probe() {

		return Probe::Get_By_Name($this->name);
	
	}

	public function Get_Stats() {

		return Stats::Get_By_Name($this->name);

	}
	
	public function Get_Iproute2_Rules() {

		$ret = array();

		if ($this->route_source_network) {
		
			$rule = new Iproute2_Rule();
			$rule->Set_Match(Rule::MATCH_SOURCE);
			$rule->Set_Value($this->network);

			$routes = array(new Iproute2_Route("default", array(new Iproute2_Gateway_Spec($this, $this->weight, $this->ip_addr))));
			
			$rule->Set_Table(Iproute2_Table::Get($routes, $this->name));
			
			$ret[] = $rule;

		}
		
		return $ret;
	
	}
	
	static public function Initialize(Config\Block $conf) {

		self::$all = array();
		
		$gws = $conf->Get_Elements_By_Name("gateway");
		
		foreach ($gws as $idx => $conf_gw) {

			$name = $conf_gw->Get_Value();

			if ($name === false) {

				throw new Exception("cannot define gateway without name");

			}
			
			$gw = new self($name);

			$gw->idx = $idx;
			$gw->conf = $conf_gw;

			$gw->ip_addr = $conf_gw->ip_addr[0];
			$gw->network = $conf_gw->network[0];
			$gw->nexthop = $conf_gw->nexthop[0];
			$gw->interface = $conf_gw->interface[0];

			if ($tmp = $conf_gw->state) {
				
				switch (strtolower($tmp[0])) {
				
					case "auto":		$state = self::STATE_AUTO;			break;
					case "active":		$state = self::STATE_ACTIVE;		break;
					case "disengaged":	$state = self::STATE_DISENGAGED;	break;
					case "disabled":	$state = self::STATE_DISABLED;		break;

					default:			throw new Exception("invalid state '{$tmp[0]}'");

				}

				$gw->admin_state = $state;
				if ($state !== self::STATE_AUTO) $gw->state = $state;

			}

			if ($tmp = $conf_gw->weight) $gw->weight = $tmp[0];

			if ($tmp = $conf_gw->probe->disengage) $gw->disengage = $tmp[0];
			if ($tmp = $conf_gw->probe->disable) $gw->disable = $tmp[0];
			if ($tmp = $conf_gw->probe->activate) $gw->activate = $tmp[0];
			if ($tmp = $conf_gw->probe->latency) $gw->max_latency = $tmp[0] / 1000;
			if ($tmp = $conf_gw->no_source_rule) $gw->route_source_network = false;
			
			self::$all[$gw->name] = $gw;
		
			if ($conf_evs = $conf_gw->events) {

				foreach ($conf_evs as $conf_ev) {

					$ev_name = $conf_ev->Get_Name();
					$ev_vals = $conf_ev->Get_Values();

					switch (strtolower($ev_name)) {

						case "activate":	$ev_type = self::EVENT_ACTIVATE;	break;
						case "disengage":	$ev_type = self::EVENT_DISENGAGE;	break;
						case "disable":		$ev_type = self::EVENT_DISABLE;		break;

						default:			throw new Exception("invalid event type '{$ev_name}'");
					
					}
					
					if (!isset($gw->events[$ev_type])) {

						$gw->events[$ev_type] = array();

					}

					foreach ($ev_vals as $ev_val) {

						$gw->events[$ev_type][] = $ev_val;

					}
			
				}
			
			}
		
		}

	}

	static public function Prepare_All() {

		foreach (self::$all as $gw) {

			$gw->Prepare();

		}
	
	}

	static public function Get_All($disabled = false) {

		$ret = array();
		
		foreach (self::$all as $gw) {

			if (($disabled) || ($gw->Get_State() !== self::STATE_DISABLED)) {

				$ret[] = $gw;

			}
		
		}
			
		return $ret;
	
	}

	static public function Get_By_Name($name) {

		foreach (self::Get_All(true) as $gw) {

			if ($gw->name === $name) {

				return $gw;

			}
		
		}
	
	}

	static public function All_Probed() {

		foreach (self::Get_All() as $gw) {

			if ($gw->status === self::STATUS_UNKNOWN) {

				return false;

			}
		
		}

		return true;
	
	}

}

class Gateway_Info {

	public $gateway;
	public $weight;

	public function __construct(Gateway $gateway, $weight = 1) {

		$this->gateway = $gateway;
		$this->weight = $weight;
	
	}

}

?>
