<?php

/**
 *
 * mpath-tools - failover and load balancing between multiple ISP made easy
 * 
 * Copyright (C) 2008-2010 Vincent Negrier aka. sIX <six at aegis-corp.org>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 *
 * @package mpath-tools
 * @subpackage iproute2
 */

namespace MPath_Tools;

use \Exception;

const IPROUTE2_DEFAULT_RT_TABLES = "/etc/iproute2/rt_tables";
const IPROUTE2_DEFAULT_RT_REALMS = "/etc/iproute2/rt_realms";
const IPROUTE2_DEFAULT_RT_PROTOS = "/etc/iproute2/rt_protos";

class Iproute2_Table_Info {

	static private $all = array();
	static private $alloc = array();

	static public $default;
	
	public $id;
	public $name;
	public $persist = false;

	static public function Initialize() {

		self::Load_All();
		self::Free_Alloc();

		if ($tmp = self::Get_By_Name("default")) {

			self::$default = $tmp;

		} else {

			self::$default = self::Alloc("default", true);

		}

	}
	
	static public function Alloc($name = NULL, $persist = false) {

		if ($table = self::Get_By_Name($name)) {

			return $table;

		}
		
		for ($i = 10; $i < 254; $i++) {

			if ((!isset(self::$all[$i])) && (!isset(self::$alloc[$i]))) {

				$ret = new self($i, isset($name) ? $name : "table_{$i}", $persist); 
				
				self::$alloc[$i] = $ret;

				return $ret;

			}

		}

		throw new Exception("cannot allocate iproute2 table (internal error)");
	
	}

	static public function Free_Alloc() {

		foreach (self::$alloc as $k => $table) {

			if (!$table->persist) {

				unset(self::$alloc[$k]);

			}
		
		}

	}
	
	static public function Get_Alloc() {

		return self::$alloc;
	
	}
	
	static public function Get_All() {

		return self::$all;

	}
	
	static public function Get_By_Name($name) {

		foreach (self::$all as $tb) {

			if ($tb->name === $name) {
				
				return $tb;

			}

		}

		return false;
	
	}

	static public function Load_All($fn = NULL) {

		if (!isset($fn)) {

			$fn = IPROUTE2_DEFAULT_RT_TABLES;

		}

		if (!$txt = @file_get_contents($fn)) {

			throw new Exception("cannot load iproute2 tables file '{$fn}'");

		}

		if (!preg_match_all("/([0-9]+)[ \t]+(.*)/", $txt, $res)) {

			throw new Exception("cannot parse iproute2 tables file '{$fn}'");

		}

		self::$all = array();
		
		foreach ($res[1] as $k => $id) {

			$name = $res[2][$k];

			if (substr($name, 0, 10) !== "mpath_dyn_") {
			
				self::$all[$id] = new self($id, $name);

			}
		
		}
	
	}
	
	public function __construct($id, $name, $persist = false) {

		$this->id = (int)$id;
		$this->name = $name;
		$this->persist = $persist;

	}

	public function Get_Flush_Command() {

		return "ip route flush table {$this->id}";
	
	}

}

class Iproute2_Realm {

	static private $all = array();
	static private $alloc = array();

	public $id;
	public $name;

	static public function Initialize() {

		self::Load_All();

	}
	
	static public function Get_All() {

		return self::$all;

	}
	
	static public function Get_By_Name($name) {

		foreach (self::$all as $tb) {

			if ($tb->name === $name) {
				
				return $tb;

			}

		}

		return false;
	
	}

	static public function Load_All($fn = NULL) {

		if (!isset($fn)) {

			$fn = IPROUTE2_DEFAULT_RT_REALMS;

		}

		if (!$txt = @file_get_contents($fn)) {

			throw new Exception("cannot load iproute2 realms file '{$fn}'");

		}

		if (!preg_match_all("/([0-9]+)[ \t]+(.*)/", $txt, $res)) {

			throw new Exception("cannot parse iproute2 realms file '{$fn}'");

		}

		self::$all = array();
		
		foreach ($res[1] as $k => $id) {

			$name = $res[2][$k];

			self::$all[$id] = new self($id, $name);
		
		}
	
	}
	
	public function __construct($id, $name) {

		$this->id = (int)$id;
		$this->name = $name;

	}

}

class Iproute2_Protocol {

	static private $all = array();
	static private $alloc = array();

	public $id;
	public $name;

	static public function Initialize() {

		self::Load_All();

	}
	
	static public function Get_All() {

		return self::$all;

	}
	
	static public function Get_By_Name($name) {

		foreach (self::$all as $tb) {

			if ($tb->name === $name) {
				
				return $tb;

			}

		}

		return false;
	
	}

	static public function Load_All($fn = NULL) {

		if (!isset($fn)) {

			$fn = IPROUTE2_DEFAULT_RT_PROTOS;

		}

		if (!$txt = @file_get_contents($fn)) {

			throw new Exception("cannot load iproute2 protocols file '{$fn}'");

		}

		if (!preg_match_all("/([0-9]+)[ \t]+(.*)/", $txt, $res)) {

			throw new Exception("cannot parse iproute2 protocols file '{$fn}'");

		}

		self::$all = array();
		
		foreach ($res[1] as $k => $id) {

			$name = $res[2][$k];

			self::$all[$id] = new self($id, $name);
		
		}
	
	}
	
	public function __construct($id, $name) {

		$this->id = (int)$id;
		$this->name = $name;

	}

}

class Iproute2_Gateway_Spec {

	public $gw;
	public $weight;
	public $realm;
	public $src;

	static private function Is_Valid_Source($src) {

		return preg_match('/^[1-9][0-9]{0,2}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}$/', $src);
	
	}
	
	public function __construct(Gateway $gw, $weight = NULL, $src = NULL, $realm = NULL) {

		$this->gw = $gw;

		if (isset($weight)) {

			$this->Set_Weight($weight);

		} else if ($gw->weight) {

			$this->Set_Weight($gw->weight);

		}
	
		if (isset($src)) $this->Set_Src($src);
		if (isset($realm)) $this->Set_Realm($realm);
	
	}

	public function Set_Weight($weight) {

		if (!is_numeric($weight)) {

			throw new Exception("weight must be numeric");

		} else if ($weight < 1) {

			throw new Exception("weight must be > 0");

		}

		$this->weight = $weight;
	
	}

	public function Set_Realm($realm) {

		$this->realm = $realm;
	
	}

	public function Set_Src($src) {

		if (!self::Is_Valid_Source($src)) {

			throw new Exception("invalid src {$src}");

		}
		
		$this->src = $src;

	}

}

class Iproute2_Route {

	public $destination;
	public $nexthops;
	public $metric = 0;
	public $proto;

	public function __construct($destination, array $nexthops, $metric = NULL, $proto = NULL) {

		$this->Set_Destination($destination);
		$this->Set_Nexthops($nexthops);
	
		if (isset($metric)) $this->Set_Metric($metric);
		if (isset($proto)) $this->Set_Proto($proto);
	
	}

	static private function Is_Valid_Destination($dest) {

		return ($dest === "default") || (preg_match('/^[1-9][0-9]{0,2}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}(\/[0-9]{1,2})?$/', $dest));
	
	}
	
	public function Set_Destination($dest) {

		if (!self::Is_Valid_Destination($dest)) {

			throw new Exception("invalid destination {$dest}");

		}
		
		$this->destination = $dest;
	
	}

	public function Set_Nexthops(array $nexthops) {

		foreach ($nexthops as $nh) {

			if ($nh instanceof Iproute2_Gateway_Spec) continue;

			throw new Exception("invalid nexthop (internal error)");

		}

		$this->nexthops = $nexthops;
	
	}

	public function Set_Metric($metric) {

		if (!is_numeric($metric)) {

			throw new Exception("metric must be numeric");

		}

		$this->metric = (int)$metric;
	
	}

	public function Set_Proto($proto) {

		$this->proto = $proto;
	
	}

}

class Iproute2_Table {

	static public $all = array();
	
	public $table_info;
	public $routes = array();

	static public function Get(array $routes, $name = NULL) {

		foreach (self::$all as $table) {

			if ($table->routes == $routes) {

				return $table;

			}

		}

		return new self(Iproute2_Table_Info::Alloc($name), $routes);
	
	}
	
	static public function Get_All() {

		return self::$all;

	}
	
	static public function Flush_All() {

		self::$all = array();
	
	}
	
	public function __construct(Iproute2_Table_Info $table, $routes = NULL)  {

		$this->table_info = $table;
		
		if (isset($routes)) {

			if (!is_array($routes)) {

				throw new Exception("routes must be an array");

			}

			foreach ($routes as $route) {

				if ($route instanceof Iproute2_Route) continue;

				throw new Exception("invalid route (internal error)");
			
			}

			$this->routes = $routes;
		
		}

		self::$all[] = $this;
	
	}

	public function Add_Route(Iproute2_Route $route) {

		$this->routes[] = $route;

	}

	public function Get_Commands() {

		$ret = array();

		$ret[] = $this->table_info->Get_Flush_Command();

		foreach ($this->routes as $route) {

			if (count($route->nexthops) === 1) {

				$via_str = "via";
				$weight = false;

			} else {

				$via_str = "nexthop via";
				$weight = true;

			}

			$tmp = array();

			foreach ($route->nexthops as $gwspec) {

				$via = "{$via_str} {$gwspec->gw->nexthop} dev {$gwspec->gw->interface}";

				if ($gwspec->weight && $weight) $via .= " weight {$gwspec->weight}";

				if ($gwspec->realm) {
					
					$via .= " realm {$gwspec->realm}";

				} else if ($realm = Iproute2_Realm::Get_By_Name($gwspec->gw->name)) {

					$via .= " realm {$realm->id}";
				
				}

				if ($gwspec->src) $via .= " src {$gwspec->src}";

				$tmp[] = $via;
			
			}

			$nexthops = implode(" ", $tmp);
			
			$cmd = "ip route add";
			
			if ($route->proto) {
				
				$cmd .= " proto {$route->proto}";

			} else if ($proto = Iproute2_Protocol::Get_By_Name("mpathd")) {

				$cmd .= " proto {$proto->id}";
			
			}

			$cmd .= " table {$this->table_info->id}";

			$cmd .= " {$route->destination} {$nexthops}";

			if ($route->metric) $cmd .= " metric {$route->metric}";

			$ret[] = $cmd;
		
		}
		
		return $ret;
	
	}

}

class Iproute2_Rule {

	static public $all = array();
	static public $main;
	
	static private $reset_idx = 10;
	static private $idx = 10;

	public $prio;
	public $match;
	public $value;

	public $table;
	public $type;

	public $collapse = true;
	
	static public function Get_All() {

		return self::$all;

	}
	
	static public function Flush_All() {

		self::$all = array();
		self::$idx = self::$reset_idx;

	}
	
	static public function Set_First_Prio($prio) {

		self::$reset_idx = $prio;
	
	}
	
	static public function Get_Last() {

		$max = 0;
		
		foreach (self::$all as $rule) {

			if ($rule->prio > $max) {

				$last_rule = $rule;
				$max = $rule->prio;

			}

		}

		return $last_rule;
	
	}

	static public function Regenerate_Source($flush = true) {

		if ($flush) {

			self::Flush_All();
			Iproute2_Table::Flush_All();
			Iproute2_Table_Info::Free_Alloc();

		}

		foreach (Gateway::Get_All(true) as $gw) {

			$gw->Get_Iproute2_Rules();

		}
	
		return self::Get_All();
	
	}
	
	static public function Regenerate_All() {

		self::Flush_All();
		Iproute2_Table::Flush_All();
		Iproute2_Table_Info::Free_Alloc();
		
		self::Regenerate_Source(false);
		
		foreach (Rule::Get_All(true) as $rule) {

			$rule->Get_Iproute2_Rules();
		
		}
	
		return self::Get_All();
	
	}
	
	static public function Renumber_All() {

		$idx = self::$reset_idx;

		foreach (self::$all as $rule) {

			$rule->Set_Prio($idx++);
		
		}

		self::$idx = $idx;
	
	}
	
	public function __construct($prio = NULL, $match = NULL, $value = NULL, $table = NULL) {

		if (isset($prio)) {
			
			$this->Set_Prio($prio);

			self::$idx = $prio + 1;

		} else {

			$this->Set_Prio(self::$idx++);

		}
		
		if (isset($match)) $this->Set_Match($match);
		if (isset($value)) $this->Set_Value($value);
		if (isset($table)) $this->Set_Table($table);

		self::$all[] = $this;

	}
	
	public function Delete() {

		foreach (self::$all as $k => $rule) {

			if ($rule === $this) {

				unset(self::$all[$k]);
				self::Renumber_All();

				return true;
			
			}
		
		}

		return false;
	
	}
	
	public function Set_Prio($prio) {

		if (!is_numeric($prio)) {

			throw new Exception("rule prio must be numeric");

		}

		$this->prio = (int)$prio;
	
	}

	public function Set_Match($match) {

		if (!in_array($match, array(Rule::MATCH_DEFAULT, Rule::MATCH_SOURCE, Rule::MATCH_DESTINATION, Rule::MATCH_FWMARK, Rule::MATCH_TOS))) {

			throw new Exception("invalid rule match {$match}");

		}

		$this->match = $match;
	
	}

	public function Set_Value($value) {

		$this->value = $value;
	
	}

	public function Set_Table(Iproute2_Table $table) {

		$this->table = $table;
	
	}

	public function Set_Type($type) {

		$this->type = $type;

	}

	public function Get_Commands() {

		$ret = array();

		$cmd = "ip rule add prio {$this->prio} ";

		switch ($this->match) {

			case Rule::MATCH_SOURCE:		$cmd .= "from {$this->value} ";									break;
			case Rule::MATCH_FWMARK:		$cmd .= "fwmark {$this->value} ";								break;
			case Rule::MATCH_TOS:			$cmd .= "tos {$this->value} ";									break;
			case Rule::MATCH_DESTINATION:	if (isset($this->type)) $cmd .= "to {$this->value} ";			break;
			case Rule::MATCH_DEFAULT:		break;

			default:						throw new Exception("unknown rule match {$this->match} (internal error)");
		
		}

		if (!isset($this->type)) {

			$cmd .= "table {$this->table->table_info->id}";
		
		} else {

			$cmd .= $this->type;

		}

		$ret[] = $cmd;
		
		return $ret;
	
	}

}

?>
