<?php

/**
 *
 * mpath-tools - failover and load balancing between multiple ISP made easy
 * 
 * Copyright (C) 2008-2010 Vincent Negrier aka. sIX <six at aegis-corp.org>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 *
 * @package mpath-tools
 * @subpackage ping
 */

namespace MPath_Tools;

use \Exception;

class Ping {
	
	const IPPROTO_ICMP = 1;
	const IPPROTO_ICMP6 = 58;
	
	const ICMP_ECHO_REQUEST = 8;
	const ICMP_ECHO_REPLY = 0;

	const ICMP6_ECHO_REQUEST = 128;
	const ICMP6_ECHO_REPLY = 129;

	const ICMP_ECHO_REQUEST_DEFAULT_SIZE = 56;
		
	private $seq = 0;
	private $streams = array();

	private $icmp_identifier;
	private $socket_family = AF_INET;

	public function __construct() {

		$this->icmp_identifier = rand(0, 0xFFFF);
	
	}
	
	public function Set_Socket_Family($af) {

		$this->socket_family = $af;

	}
	
	static private function Calc_Checksum($data) {

		 $bit = unpack("n*", $data);
		 $sum = array_sum($bit);

		 if (strlen($data) % 2) {

			 $temp = unpack("C*", $data[strlen($data) - 1]);
			 $sum += $temp[1];

		 }

		 $sum = ($sum >> 16) + ($sum & 0xffff);
		 $sum += ($sum >> 16);

	     return pack("n*", ~$sum);

	}
	
	static private function Build_ICMP_Packet($type, $code, $payload) {

		$hdr = chr($type) . chr($code);

		return $hdr . self::Calc_Checksum($hdr . "\x00\x00" . $payload) . $payload;

	}

	private function Inc_Sequence() {

		$this->seq = ($this->seq + 1) & 0xFFFF;

		return $this->seq;
	
	}
	
	private function Build_ICMP_Echo_Request_Packet($size, $pattern, $seq = NULL) {

		if (!isset($seq)) {

			$seq = $this->Inc_Sequence();

		} else {

			$this->seq = $seq;

		}
		
		$msg  = pack("n", $this->icmp_identifier);
		$msg .= pack("n", $seq);
		$msg .= str_pad("", $size, $pattern);

		switch ($this->socket_family) {

			case AF_INET:		$type = self::ICMP_ECHO_REQUEST;	break;
			case AF_INET6:		$type = self::ICMP6_ECHO_REQUEST;	break;

		}
		
		return self::Build_ICMP_Packet($type, 0, $msg);

	}

	public function Send_ICMP_Echo_Request($addr, array $source_addrs = array(), $size = self::ICMP_ECHO_REQUEST_DEFAULT_SIZE, $pattern = "\x55\xAA", $seq = NULL, $throw_on_error = true) {

		$ret = array();
		
		$pkt = $this->Build_ICMP_Echo_Request_Packet($size, $pattern, $seq);
		$pktlen = strlen($pkt);

		switch ($this->socket_family) {

			case AF_INET:		$def_source_addrs = array("localhost" => "127.0.0.1");		
								$ipproto = self::IPPROTO_ICMP;
								break;

			case AF_INET6:		$def_source_addrs = array("ip6-localhost" => "::1");		
								$ipproto = self::IPPROTO_ICMP6;
								break;
		
		}

		
		if (!$source_addrs) {

			$source_addrs = $def_source_addrs;
		
		}
		
		foreach ($source_addrs as $source_id => $source_addr) {

			$err = false;
			
			if (!isset($this->streams[$source_id])) {

				if (!$sck = socket_create($this->socket_family, SOCK_RAW, $ipproto)) {

					$err = "socket_create(): " . socket_strerror(socket_last_error());

				} else if (!socket_bind($sck, $source_addr)) {

					$err = "socket_bind({$source_addr}): " . socket_strerror(socket_last_error($sck));

				} else if (!socket_connect($sck, $addr, NULL)) {

					$err = "socket_connect({$addr}): " . socket_strerror(socket_last_error($sck));
				
				} else if (!socket_set_nonblock($sck)) {

					$err = "socket_set_nonblock(): " . socket_strerror(socket_last_error($sck));
				
				} else {

					$this->streams[$source_id] = $sck;

				}
			
			}

			if ($err === false) {
			
				$mt = microtime(true);
				
				if (!socket_send($this->streams[$source_id], $pkt, $pktlen, 0)) {

					$err = "socket_send(): " . socket_strerror(socket_last_error($this->streams[$source_id]));
				
				}

			}
			
			if ($err === false) {
			
				$ret[$source_id] = $mt;

			} else if ($throw_on_error) {

				throw new Exception($err);

			} else {
				
				$ret[$source_id] = $err;

			}

		}
	
		return $ret;
	
	}

	public function Has_Waiting_Data($timeout = 0) {

		$rfd = $this->streams;
		$dfd = array();

		$tm_sec = (int)$timeout;
		$tm_usec = (int)(($timeout - $tm_sec) * 1000000);
		
		if (socket_select($rfd, $dfd, $dfd, $tm_sec, $tm_usec)) {

			return $rfd;

		} else {

			return false;

		}
	
	}

	static public function Poll_Multi(array $instances, $timeout = 0) {

		$rfd = $dfd = array();
		$inst_idx = array();

		$i = 0;
		
		foreach ($instances as $k => $mp) {

			foreach ($mp->streams as $fd) {

				$rfd[$i] = $fd;
				$inst_idx[$i] = array($k, $mp);

				++$i;
			
			}
		
		}
	
		$tm_sec = (int)$timeout;
		$tm_usec = (int)(($timeout - $tm_sec) * 1000000);
		
		if (socket_select($rfd, $dfd, $dfd, $tm_sec, $tm_usec)) {

			$ret = array();

			foreach ($rfd as $k => $fd) {

				$ret[$inst_idx[$k][0]] = $inst_idx[$k][1];
			
			}

			return $ret;

		} else {

			return false;

		}
	
	}
	
	public function Get_Responses($timeout = 0) {

		if (!$rfd = $this->Has_Waiting_Data($timeout)) {

			return false;

		}

		$ret = array();

		foreach ($rfd as $source_id => $fd) {

			$read = socket_recvfrom($fd, $pkt, 65536, 0, $src, $port);
						
			switch ($this->socket_family) {

				case AF_INET:	$reply = ICMP_Echo_Reply::Decode($src, $pkt, $this->icmp_identifier);		break;
				case AF_INET6:	$reply = ICMP6_Echo_Reply::Decode($src, $pkt, $this->icmp_identifier);		break;

			}
			
			if ($reply) {

				$ret[$source_id][] = $reply;

			}
		
		}
	
		return $ret;
	
	}

	public function Do_Ping($addr, array $source_addrs = array(), $size = self::ICMP_ECHO_REQUEST_DEFAULT_SIZE, $pattern = "\x55\xAA", $timeout = 4, $callback = NULL, $seq = NULL) {

		$ret = $callbacks = array();
		
		$sent_mt = $this->Send_ICMP_Echo_Request($addr, $source_addrs, $size, $pattern, $seq, false);

		$wait_resp = 0;
		
		foreach ($sent_mt as $source_id => $val) {

			if (is_string($val)) {
				
				$callbacks[] = array(false, NULL, $source_id, $val, NULL);
			
			} else {
				
				$wait_resp++;

			}
		
		}
		
		$expect = str_pad("", $size, $pattern);
		
		if ($callbacks && is_callable($callback)) {
			
			foreach ($callbacks as $cb_params) {
				
				call_user_func_array($callback, $cb_params);

			}

			$callbacks = array();

		}
	
		do {

			$start_mt = microtime(true);

			$rfd = $this->streams;
			$dfd = array();

			$tm_sec = (int)$timeout;
			$tm_usec = (int)(($timeout - $tm_sec) * 1000000);

			if (socket_select($rfd, $dfd, $dfd, $tm_sec, $tm_usec)) {

				foreach ($rfd as $source_id => $fd) {

					$pkt = "";
					$src = "";
					$port = 0;
					
					$read = socket_recvfrom($fd, $pkt, 65536, 0, $src, $port);

					if ($read !== false) {
					
						switch ($this->socket_family) {

							case AF_INET:	$reply = ICMP_Echo_Reply::Decode($src, $pkt, $this->icmp_identifier);		break;
							case AF_INET6:	$reply = ICMP6_Echo_Reply::Decode($src, $pkt, $this->icmp_identifier);		break;

						}

						if ($reply) {

							if ($reply->seq === $this->seq) {
							
								if (strlen($reply->data) === $size) {
								
									if ($reply->data === $expect) {

										if (!isset($ret[$source_id])) {
									
											$ret[$source_id] = microtime(true) - $sent_mt[$source_id];

											$callbacks[] = array(true, $src, $source_id, $ret[$source_id], $reply);

											$wait_resp--;

										} else {

											$callbacks[] = array(false, $src, $source_id, "duplicate reply with icmp_seq={$reply->seq}", $reply);
										
										}
									
									} else {

										$callbacks[] = array(false, $src, $source_id, "wrong payload data", $reply);
									
									}
							
								} else {

									$callbacks[] = array(false, $src, $source_id, "wrong size (expected {$size}, got " . strlen($reply->data) . ")", $reply);
								
								}
							
							} else {

								$callbacks[] = array(false, $src, $source_id, "wrong sequence number (expected {$this->seq}, got {$reply->seq})", $reply);

							}
						
						}

					} else {

						$callbacks[] = array(false, $src, $source_id, "socket error: " . socket_strerror(socket_last_error($fd)), $reply);
					
					}

				}

				if (is_callable($callback)) {
					
					foreach ($callbacks as $cb_params) {
						
						call_user_func_array($callback, $cb_params);

					}

					$callbacks = array();

				}
			
			}
		
			pcntl_signal_dispatch();
			
			$timeout -= (microtime(true) - $start_mt);

		} while ($wait_resp && ($timeout > 0));

		return $ret;
	
	}

	public function __destruct() {

		foreach ($this->streams as $fd) {

			socket_close($fd);
		
		}
	
	}

}

class ICMP_Echo_Reply {

	public $addr;
	public $type;
	public $code;
	public $cksum;
	public $id;
	public $seq;
	public $ttl;
	public $data;

	static public function Decode($addr, $pkt, $icmp_identifier) {

		$ttl = ord($pkt[8]);
		$pkt = substr($pkt, 20);
		$tmp = unpack("Ctype/Ccode/ncksum/nid/nseq", $pkt);

		if (($tmp["type"] === Ping::ICMP_ECHO_REPLY) && ($tmp["code"] === 0) && ($tmp["id"] === $icmp_identifier)) {
		
			$ret = new self();
			
			$ret->addr = $addr;
			
			$ret->type = $tmp["type"];
			$ret->code = $tmp["code"];
			$ret->cksum = $tmp["cksum"];
			$ret->id = $tmp["id"];
			$ret->seq = $tmp["seq"];
			
			$ret->data = substr($pkt, 8);
			$ret->ttl = $ttl;

			return $ret;

		} else {

			return false;

		}
	
	}

}

class ICMP6_Echo_Reply extends ICMP_Echo_Reply {

	static public function Decode($addr, $pkt, $icmp_identifier) {

		$ttl = ord($pkt[8]);
//		$pkt = substr($pkt, 20);
		$tmp = unpack("Ctype/Ccode/ncksum/nid/nseq", $pkt);

		if (($tmp["type"] === Ping::ICMP6_ECHO_REPLY) && ($tmp["code"] === 0) && ($tmp["id"] === $icmp_identifier)) {
		
			$ret = new self();
			
			$ret->addr = $addr;
			
			$ret->type = $tmp["type"];
			$ret->code = $tmp["code"];
			$ret->cksum = $tmp["cksum"];
			$ret->id = $tmp["id"];
			$ret->seq = $tmp["seq"];
			
			$ret->data = substr($pkt, 8);
			$ret->ttl = $ttl;

			return $ret;

		} else {

			return false;

		}
	
	}

}

?>