<?php

/**
 *
 * mpath-tools - failover and load balancing between multiple ISP made easy
 * 
 * Copyright (C) 2008-2010 Vincent Negrier aka. sIX <six at aegis-corp.org>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 *
 * @package mpath-tools
 * @subpackage probe
 */

namespace MPath_Tools;

use \Exception;

class Probe {
	
	const DEFAULT_INTERVAL = 2;
	const DEFAULT_TIMEOUT = 3;
	
	const STATE_IDLE = 1;
	const STATE_WAIT = 2;
	
	const STATUS_UNKNOWN = 0;
	const STATUS_OK = 1;
	const STATUS_TIMEOUT = 2;
	const STATUS_ERROR = 3;
	
	static private $all = array();
	
	public $gateway;
	
	private $gwname;
	
	private $addrs = array();
	private $last_sent_mt = 0;
	private $last_recv_mt = 0;
	private $wait_addrs = array();
	
	private $state = self::STATE_IDLE;
	
	private $status = self::STATUS_UNKNOWN;
	private $status_change_mt = 0;

	public $latency = 0;

	public $interval = self::DEFAULT_INTERVAL;
	public $timeout = self::DEFAULT_TIMEOUT;
	
	private $ping;
	private $log = array();
	
	static public function Initialize() {

		self::$all = array();
		
		foreach (Gateway::Get_All() as $gw) {

			self::$all[$gw->name] = new self($gw);

		}
	
	}

	static public function Get_By_Name($gwname) {

		if (!isset(self::$all[$gwname])) {

			throw new Exception("no probe for gateway '{$gwname}' (internal error)");
		
		}
		
		return self::$all[$gwname];
		
	}
	
	private function __construct(Gateway $gw) {
		
		$this->gateway = $gw;

		$this->gwname = $gw->name;

		if ($addrs = $gw->conf->probe->ip_addr) {
			
			$this->addrs = $addrs->Get_Values();
			
		} else {
			
			$this->addrs = array($gw->nexthop);

		}

		if ($iv = $gw->conf->probe->interval) {
			
			$this->interval = $iv->Get_Value();

		}

		if ($tm = $gw->conf->probe->timeout) {
			
			$this->timeout = $tm->Get_Value();

		}

		self::$all[$gw->name] = $this;
	
	}

	private function Log($status = NULL, $latency = NULL) {

		if (!isset($status)) {

			$status = $this->status;

		}
		
		$log_entry = array($status);

		if ($status === self::STATUS_OK) {

			if (!isset($latency)) {

				$latency = $this->latency;

			}
			
			$log_entry[] = $latency;

		}

		$this->log[] = $log_entry;
	
	}

	public function Get_Log() {

		$ret =$this->log;

		$this->log = array();

		return $ret;
	
	}
	
	private function Send() {

		try {
		
			$this->ping = new Ping();

			$src = array($this->gateway->ip_addr);

			foreach ($this->addrs as $addr) {

				$this->ping->Send_ICMP_Echo_Request($addr, $src);

			}
			
			$this->wait_addrs = $this->addrs;
			$this->last_sent_mt = microtime(true);
			$this->state = self::STATE_WAIT;
	
		} catch (Exception $e) {

			$this->status = self::STATUS_ERROR;
			$this->last_recv_mt = microtime(true);
			$this->state = self::STATE_IDLE;

			$this->Log();

		}
	
	}

	private function Receive() {

		if (!$resp = $this->ping->Get_Responses()) {

			return false;

		}

		foreach ($resp as $src => $replies) {

			foreach ($replies as $reply) {
			
				if (in_array($reply->addr, $this->wait_addrs)) {

					$this->status = self::STATUS_OK;
					$this->last_recv_mt = microtime(true);
					$this->latency = $this->last_recv_mt - $this->last_sent_mt;

					$this->Log();

					unset($this->wait_addrs[array_search($reply->addr, $this->wait_addrs)]);

					if (!$this->wait_addrs) {

						$this->state = self::STATE_IDLE;

					}
				
				}
		
			}
		
		}
	
	}

	static public function Run() {

		$mt = microtime(true);
		
		$pings = array();
		
		foreach (self::$all as $k => $probe) {

			switch ($probe->state) {

				case self::STATE_IDLE:

				if ($mt > ($probe->last_recv_mt + $probe->interval)) {
					
					$probe->Send();
					$pings[$k] = $probe->ping;

				}
				
				break;
				
				
				case self::STATE_WAIT:

				if ($mt > ($probe->last_sent_mt + $probe->timeout)) {

					$probe->status = (count($probe->wait_addrs) === count($probe->addrs)) ? self::STATUS_TIMEOUT : self::STATUS_OK;
					$probe->state = self::STATE_IDLE;
					$probe->Log();

				} else {
				
					$pings[$k] = $probe->ping;

				}
				
				break;
			
			}
		
		}

		if ($pings) {

			if ($replies = Ping::Poll_Multi($pings, 1)) {

				foreach ($replies as $k => $ping) {

					self::$all[$k]->Receive();

				}
			
			}
		
		}
	
		$ret = array();

		foreach (self::$all as $probe) {

			$ret[$probe->gateway->name] = array($probe->status, $probe->latency, $probe->Get_Log());
		
		}

		return $ret;
	
	}

	static public function Has_Waiting() {

		foreach (self::$all as $probe) {

			if ($probe->state === self::STATE_WAIT) {

				return true;

			}

		}

		return false;

	}

	static public function Sleep_Until_Next_Probe() {

		$next_probe = PHP_INT_MAX;
		
		foreach (self::$all as $probe) {

			if (($probe->state === self::STATE_WAIT) || (!$probe->last_recv_mt)) {

				return false;

			}

			$tmp = $probe->last_recv_mt + $probe->interval;

			if ($tmp < $next_probe) {

				$next_probe = $tmp;

			}
		
		}

		$st = $next_probe - microtime(true);

		if ($st <= 0) {
			
			return false;
			
		} else if ($st > 10) {
			
			$st = 10;

		}

		usleep($st * 1000000);

	}

}

?>
