<?php

/**
 *
 * mpath-tools - failover and load balancing between multiple ISP made easy
 * 
 * Copyright (C) 2008-2010 Vincent Negrier aka. sIX <six at aegis-corp.org>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 *
 * @package mpath-tools
 * @subpackage rule
 */

namespace MPath_Tools;

use \Exception;

class Rule {

	const GATEWAY_WEIGHT_SEPARATOR = ":";
	
	const MATCH_DEFAULT = 0;
	const MATCH_SOURCE = 1;
	const MATCH_DESTINATION = 2;
	const MATCH_FWMARK = 3;
	const MATCH_TOS = 4;
	
	const DEF_DEFAULT = 0;
	const DEF_UNREACHABLE = 1;
	const DEF_BLACKHOLE = 2;
	const DEF_PROHIBIT = 3;
	
	static public $all = array();
	static public $default;
	
	static private $uid = 1;
	
	public $id;
	public $match;
	public $conf_value = array();
	public $value = array();
	public $default_policy = self::DEF_DEFAULT;
	public $expires = false;
	public $routes = array();
	public $table;

	static private function Create_From_Conf_Element(Config\Block $conf_rule, $do_resolve = true) {

		$match = self::Str_To_Match($conf_rule->Get_Name());
		$value = $conf_rule->Get_Values();
		
		$routes = array();
		
		foreach ($conf_rule->Get_Elements_By_Name("via") as $via) {

			$routes[] = $via->Get_Values();

		}

		if ($conf_rule->policy) {
		
			switch (strtolower($conf_rule->policy)) {

				case "unreachable":	$default_policy = self::DEF_UNREACHABLE;	break;
				case "blackhole":	$default_policy = self::DEF_BLACKHOLE;		break;
				case "prohibit":	$default_policy = self::DEF_PROHIBIT;		break;

				default:			throw new Exception("unknown default policy '{$conf_rule->policy}'");
			
			}
		
		} else {

			$default_policy = self::DEF_DEFAULT;

		}
		
		$expires = false;
		
		if ($tmp = $conf_rule->table) $table = $tmp[0];
		
		return new self($routes, $match, $value, $default_policy, $expires, $do_resolve, $table);
	
	}

	static public function Create_From_Array(array $rule) {

		return new self($rule["via"], $rule["match"], $rule["value"], isset($rule["default_policy"]) ? $rule["default_policy"] : Rule::DEF_DEFAULT, isset($rule["expires"]) ? $rule["expires"] : false);

	}
	
	static public function Str_To_Policy($str) {

		switch (strtolower($str)) {

			case "default":					return self::DEF_DEFAULT;
			case "unreachable":				return self::DEF_UNREACHABLE;
			case "blackhole":				return self::DEF_BLACKHOLE;
			case "prohibit":				return self::DEF_PROHIBIT;

		}

		throw new Exception("invalid policy '{$str}'");
	
	}
	
	static public function Policy_To_Str($policy) {

		switch ($policy) {

			case self::DEF_DEFAULT:			return "default";
			case self::DEF_UNREACHABLE:		return "unreachable";
			case self::DEF_BLACKHOLE:		return "blackhole";
			case self::DEF_PROHIBIT:		return "prohibit";

			default:						return "??";
		
		}
	
	}

	static public function Str_To_Match($str) {

		switch (strtolower($str)) {

			case "default":					return self::MATCH_DEFAULT;
			case "source":					return self::MATCH_SOURCE;
			case "destination":				return self::MATCH_DESTINATION;
			case "fwmark":					return self::MATCH_FWMARK;
			case "tos":						return self::MATCH_TOS;
		
		}

		throw new Exception("invalid match type '{$str}'");
	
	}
	
	static public function Match_To_Str($match) {

		switch ($match) {

			case self::MATCH_DEFAULT:		return "default";
			case self::MATCH_SOURCE:		return "source";
			case self::MATCH_DESTINATION:	return "destination";
			case self::MATCH_FWMARK:		return "fwmark";
			case self::MATCH_TOS:			return "tos";

			default:						return "??";

		}
	
	}
	
	static public function Initialize(Config\Block $conf, $do_resolve = true) {

		self::$all = array();
		self::$default = NULL;
		
		if ($conf->rules) {
		
			try {
			
				foreach ($conf->rules as $crule) {

					self::Create_From_Conf_Element($crule, $do_resolve);
				
				}

			} catch (Exception $e) {

				throw new Exception("cannot initialize rules: {$e->getMessage()}");

			}
		
		}

		if (!isset(self::$default)) {

			$all_gws = Gateway::Get_All();
			$all_gw_names = array();

			foreach ($all_gws as $gw) {

				$all_gw_names[] = $gw->name;

			}

			self::$default = new self($all_gw_names, self::MATCH_DEFAULT);
		
		}
	
	}
	
	static public function Refresh_All() {

		$ret = 0;
		
		foreach (self::Get_All() as $rule) {

			$ret |= (int)$rule->Resolve_Hostnames();

		}

		return (bool)$ret;
	
	}
	
	static public function Get_All($include_default = false) {

		$ret = self::$all;

		if ($include_default) $ret[] = self::$default;

		return $ret;

	}

	static public function Exists($match, $value = array()) {

		try {

			Rule::Get($match, $value);

			return true;
		
		} catch (Exception $e) {

			return false;

		}

	}
		
	static public function Get($match, $value = array()) {

		if ($match === self::MATCH_DEFAULT) {

			if (isset(self::$default)) {

				return self::$default;

			} else {

				throw new Exception("no default rule");

			}

		}
		
		foreach (self::Get_All() as $rule) {

			if (($rule->match === $match) && ($rule->conf_value === $value)) {

				return $rule;

			}
		
		}

		throw new Exception("rule not found: " . self::Match_To_Str($match) . " " . implode(" ", $value));
	
	}

	static public function Do_Maintenance($t = NULL) {

		$ret = 0;
		
		if (!isset($t)) {
		
			$t = time();

		}
		
		foreach (self::Get_All() as $rule) {

			if (($rule->expires !== false) && ($rule->expires <= $t)) {

				$rule->Remove();

				$ret++;

			}
		
		}

		return $ret;
	
	}
	
	public function __construct(array $routes, $match, array $value = array(), $default = NULL, $expires = false, $do_resolve = true, $table = NULL) {

		if (self::Exists($match, $value)) {

			throw new Exception("duplicate rule: " . self::Match_To_Str($match) . " " . implode(" ", $value));

		}
		
		$this->Set_Match($match);
		$this->Set_Value($value, $do_resolve);
		$this->Set_Expires($expires);

		if (isset($default)) {

			$this->Set_Default_Policy($default);

		}

		$this->Set_Routes($routes);
		
		if (isset($table)) {

			$this->Set_Table($table);

		}
		
		if ($this->match === self::MATCH_DEFAULT) {

			self::$default = $this;

		} else {

			$this->id = self::$uid++;
			
			self::$all[$this->id] = $this;

		}
	
	}

	public function Set_Match($match) {

		if (!in_array($match, array(	self::MATCH_DEFAULT,
										self::MATCH_SOURCE,
										self::MATCH_DESTINATION,
										self::MATCH_FWMARK,
										self::MATCH_TOS))) {

			throw new Exception("invalid match {$match}");
		
		}

		$this->match = $match;
	
	}

	public function Set_Value($value, $do_resolve = true) {

		if (($this->match !== self::MATCH_DEFAULT) && (!is_array($value))) {

			throw new Exception("invalid match value {$value}");

		}

		$this->conf_value = $this->value = $value;

		if ($do_resolve) {
		
			$this->Resolve_Hostnames();

		}

	}

	public function Set_Default_Policy($policy) {

		if (!in_array($policy, array(	self::DEF_DEFAULT,
										self::DEF_UNREACHABLE,
										self::DEF_BLACKHOLE,
										self::DEF_PROHIBIT))) {

			throw new Exception("invalid default policy {$policy}");
		
		}

		$this->default_policy = $policy;
	
	}

	public function Set_Expires($expires) {

		if ($expires === false) {

			$this->expires = false;

		} else if (is_numeric($expires)) {

			if ($this->match === self::MATCH_DEFAULT) {

				throw new Exception("default rule cannot expire");
			
			}
			
			$this->expires = (int)$expires;

		} else {
		
			throw new Exception("invalid expire value {$expires}");

		}

	}

	public function Set_Routes(array $routes) {

		if (count($routes) === 0) {

			throw new Exception("there must be at least one gateway");

		}
		
		foreach ($routes as $route) {

			if (!is_array($route) || (count($route) === 0)) {

				throw new Exception("invalid route value {$route}");
			
			}
			
			foreach ($route as $gw_info) {

				$gw = strtok($gw_info, self::GATEWAY_WEIGHT_SEPARATOR);
				
				if (!Gateway::Get_By_Name($gw)) {

					throw new Exception("invalid gateway name '{$gw}'");
				
				}
			
			}
		
		}

		$this->routes = $routes;
	
	}
	
	public function Set_Table($table) {

		$this->table = $table;
	
	}
	
	public function Reorder($idx) {
		
		if ($this->match === self::MATCH_DEFAULT) {

			throw new Exception("cannot change default rule position");

		}
		
		$i = 0;
		$tmp = array();
		$done = false;

		foreach (self::$all as $k => $r) {

			if (++$i == $idx) {

				$tmp[$this->id] = $this;
				$done = true;

			}
		
			if ($r !== $this) {
			
				$tmp[$k] = $r;

			}
		
		}

		if (!$done) {

			$tmp[$this->id] = $this;

		}
		
		self::$all = $tmp;
	
		if (!$done) {

			throw new Exception("invalid rule position {$idx}");

		}
	
	}
	
	public function Remove() {

		if ($this->match === self::MATCH_DEFAULT) {

			throw new Exception("cannot remove default rule");

		}

		unset(self::$all[$this->id]);
	
	}
	
	private function Resolve_Hostnames() {

		$num_hostnames = $num_ips = 0;
		
		if (($this->match !== self::MATCH_SOURCE) && ($this->match !== self::MATCH_DESTINATION)) {

			return false;

		}
		
		$new_value = array();
		
		foreach ($this->conf_value as $tmp) {

			if (preg_match("/[1-9][0-9]{0,2}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}(?:\/[0-9]{1,2})/", $tmp)) {

				$new_value[] = $tmp;
			
			} else {
			
				$num_hostnames++;
				
				$tmp2 = gethostbynamel($tmp);

				if (is_array($tmp2)) {

					foreach ($tmp2 as $addr) {

						$num_ips++;
						
						$new_value[] = $addr;

					}
				
				}

			}
		
		}
	
		if ($this->value !== $new_value) {

			$this->value = $new_value;

			return true;
		
		} else {
		
			return false;

		}
	
	}

	public function Pick_Gateways(array $gws, $disengaged = false) {

		foreach ($this->routes as $route) {
		
			$route_gws = $route_weights = array();
			
			foreach ($route as $gw_info) {

				$gwname = strtok($gw_info, self::GATEWAY_WEIGHT_SEPARATOR);
				$weight = strtok("");

				$route_gws[] = $gwname;
				$route_weights[] = ($weight ? $weight : 1);

			}
			
			$ret = array();
			
			foreach ($gws as $gw) {

				if ((in_array($gw->name, $route_gws)) && ($gw->Get_State() === ($disengaged ? Gateway::STATE_DISENGAGED : Gateway::STATE_ACTIVE))) {

					$ret[] = new Gateway_Info($gw, $route_weights[array_search($gw->name, $route_gws)]);
					
				}
			
			}

			if ($ret) {

				break;

			}
	
		}

		if (!$ret && !$disengaged) {

			$ret = self::Pick_Gateways($gws, true);

		}
	
		return $ret;
	
	}

	public function Get_Iproute2_Rules($all_gws = NULL) {

		$use_gws = $this->Pick_Gateways(isset($all_gws) ? $all_gws : Gateway::Get_All(true));

		$gw_specs = array();
		
		foreach ($use_gws as $gw) {

			$weight = $gw->weight
			or $weight = $gw->gateway->weight;

			$gw_specs[] = new Iproute2_Gateway_Spec($gw->gateway, $weight);

		}

		$ret = array();

		if ($this->match === self::MATCH_DEFAULT) {

			$rule = new Iproute2_Rule();
			$rule->Set_Match(self::MATCH_DEFAULT);

			if ($use_gws) {

				$routes = array(new Iproute2_Route("default", $gw_specs));

				$i = 1;
				
				foreach ($gw_specs as $gwspec) {

					$routes[] = new Iproute2_Route("default", array($gwspec), $i++);
				
				}

				if ($rule->table) {

					foreach ($routes as $route) {

						$rule->table->Add_Route($route);

					}

				} else {
				
					$rule->Set_Table(Iproute2_Table::Get($routes, $this->table));

				}

				$ret[] = $rule;
			
			} else if ($this->default_policy !== self::DEF_DEFAULT) {

				$rule->Set_Type(self::Policy_To_Str($this->default_policy));
			
				$ret[] = $rule;
			
			} else {

				$rule->Delete();

			}
		
		} else if ($this->match === self::MATCH_DESTINATION) {

			if (($use_gws) && ($last = Iproute2_Rule::Get_Last()) && ($last->match === self::MATCH_DEFAULT) && (!isset($this->table)) && ($last->collapse)) {

				$rule = $last;

			} else {
			
				$rule = new Iproute2_Rule();
				$rule->Set_Match(self::MATCH_DEFAULT);

			}
			
			$rule->collapse = !isset($this->table);
			
			if ($use_gws) {

				$routes = array();
				
				foreach ($this->value as $val) {

					$routes[] = new Iproute2_Route($val, $gw_specs);

				}
				
				if ($rule->table) {

					foreach ($routes as $route) {
					
						$rule->table->Add_Route($route);

					}

				} else {
				
					$rule->Set_Table(Iproute2_Table::Get($routes, $this->table));

				}
			
			} else {
				
				if ($this->default_policy !== self::DEF_DEFAULT) {

					foreach ($this->value as $val) {
					
						$prule = new Iproute2_Rule();
						$prule->Set_Match(self::MATCH_DESTINATION);
						$prule->Set_Value($val);
						$prule->Set_Type(self::Policy_To_Str($this->default_policy));

					}

				}
			
				$rule->Delete();
			
			}
		
			if (!isset($prule)) {
				
				$ret[] = $rule;

			}
		
		} else {
		
			foreach ($this->value as $val) {

				$rule = new Iproute2_Rule();
				$rule->Set_Match($this->match);
				$rule->Set_Value($val);
				
				if ($use_gws) {

					$def_route = new Iproute2_Route("default", $gw_specs);

					$rule->Set_Table(Iproute2_Table::Get(array($def_route), $this->table));
				
					$ret[] = $rule;
				
				} else if ($this->default_policy !== self::DEF_DEFAULT) {

					$rule->Set_Type(self::Policy_To_Str($this->default_policy));
				
					$ret[] = $rule;
				
				} else {

					$rule->Delete();

				}
			
			}
		
		}
		
		return $ret;
	
	}

}

?>