<?php

/**
 *
 * mpath-tools - failover and load balancing between multiple ISP made easy
 * 
 * Copyright (C) 2008-2010 Vincent Negrier aka. sIX <six at aegis-corp.org>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 *
 * @package mpath-tools
 * @subpackage stats
 */

namespace MPath_Tools;

use \Exception;

class Stats {

	const MAX_AGE = 3600;
	
	static private $all = array();
	
	private $nb_probes = 0;
	private $nb_status = array();

	private $last_data = array();

	private function __construct() {

	}

	static public function Get_By_Name($gwname) {

		if (!isset(self::$all[$gwname])) {

			throw new Exception("no stats for gateway '{$gwname}' (internal error)");
		
		}
		
		return self::$all[$gwname];
	
	}
	
	static public function Add_Gateway_Data($gwname, $status, $latency, $t = false) {

		if (!isset(self::$all[$gwname])) {

			self::$all[$gwname] = new self();

		}

		return self::$all[$gwname]->Add_Data($status, $latency, $t);
	
	}
	
	public function Add_Data($status, $latency, $t = false) {

		if ($t === false) {

			$t = time();

		}
		
		$this->nb_probes++;
		$this->nb_status[$status]++;
		
		$this->last_data[] = array($t, $status, $latency);

		$this->Prune();
	
	}

	public function Prune($max_age = NULL, $t = NULL) {

		if (!isset($t)) {

			$t = time();

		}
		
		if (!isset($max_age)) {

			$min_t = $t - self::MAX_AGE;

		}

		foreach ($this->last_data as $k => $tmp) {

			if ($tmp[0] < $min_t) {

				unset($this->last_data[$k]);

			} else {

				break;

			}
		
		}
	
	}
	
	private function Get_Samples($last_seconds = NULL, $only_ok = false) {

		if (!isset($last_seconds)) {

			$ret = $this->last_data;

		} else {

			$ret = array();
			
			$min_time = time() - $last_seconds;
			
			foreach (array_reverse($this->last_data) as $tmp) {
				
				if ($tmp[0] < $min_time) return $ret;

				if ($only_ok) {
					
					if ($tmp[1] !== Probe::STATUS_OK) continue;

					$ret[] = $tmp[2];
				
				} else {

					$ret[] = $tmp;

				}

			}
		
		}
	
		return $ret;
	
	}
	
	public function Get_Probe_Data($last_seconds = 300) {

		$tot_smp = $tot_sent = $tot_rcvd = $tmo_smp = $err_smp = 0;

		foreach ($this->Get_Samples($last_seconds) as $tmp) {
			
			$tot_sent++;
			
			switch ($tmp[1]) {

				case Probe::STATUS_UNKNOWN:					continue 2;
				case Probe::STATUS_OK:		$tot_rcvd++;	break;
				case Probe::STATUS_TIMEOUT:	$tmo_smp++;		break;
				case Probe::STATUS_ERROR:	$err_smp++;		$tot_sent--;	break;

				default:					throw new Exception("unknown probe status '{$tmp[1]}' (internal error)");

			}

			$tot_smp++;
			
		}

		return array(	"sent"		=> $tot_sent,
						"received"	=> $tot_rcvd,
						"error"		=> $err_smp,
						"timeout"	=> $tmo_smp,
						"% loss"    => ($tmo_smp / $tot_smp) * 100,
						"% error"	=> ($err_smp / $tot_smp) * 100);
		
	}
	
	public function Get_Packet_Loss_Pct($last_seconds = 300) {

		$tmp = $this->Get_Probe_Data($last_seconds);

		return $tmp["% loss"] + $tmp["% error"];
		
	}

	public function Get_Rtt_Data($last_seconds = 300) {

		$tot = 0;
		
		$rtts = $this->Get_Samples($last_seconds, true);
		$rcvd = count($rtts);

		foreach ($rtts as $time) {

			$tot += $time;

			if ((!isset($rtt_min)) || ($time < $rtt_min)) {

				$rtt_min = $time;

			}
		
			if ((!isset($rtt_max)) || ($time > $rtt_max)) {

				$rtt_max = $time;

			}
		
		}
		
		if ($rcvd > 1) {
			
			$rtt_avg = $tot / $rcvd;

			$rtt_mdev = 0;
			
			foreach ($rtts as $time) {

				$rtt_mdev += pow($time - $rtt_avg, 2);
				
			}

			$rtt_mdev = sqrt($rtt_mdev / ($rcvd - 1));

		} else {

			$rtt_avg = $tot;
			$rtt_mdev = 0;

		}
			
		return array(	"rtt min"	=> $rtt_min * 1000,
						"rtt avg"	=> $rtt_avg * 1000,
						"rtt max"	=> $rtt_max * 1000,
						"rtt mdev"	=> $rtt_mdev * 1000);
	
	}

}

?>