#!/usr/local/bin/php
<?php

/**
 *
 * mpath-tools - failover and load balancing between multiple ISP made easy
 * 
 * Copyright (C) 2008-2010 Vincent Negrier aka. sIX <six at aegis-corp.org>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 *
 * @package mpath-tools
 * @subpackage mpathd
 */

namespace MPath_Tools;

const DEFAULT_CONF_FILENAME = "mpathd.conf";

require "mpath-tools/inc/nanoserv/nanoserv.php";
require "mpath-tools/inc/config.php";
require "mpath-tools/inc/gateway.php";
require "mpath-tools/inc/rule.php";
require "mpath-tools/inc/ping.php";
require "mpath-tools/inc/probe.php";
require "mpath-tools/inc/log.php";
require "mpath-tools/inc/iproute2.php";
require "mpath-tools/inc/stats.php";
require "mpath-tools/inc/control.php";
require "mpath-tools/inc/control_server.php";

use \Exception;
use \Tokenizer\Exception as Tokenizer_Exception;

use \MPath_Tools\Config\Core as Conf;
use \MPath_Tools\Control\Server_Connection as Control_Server;
use \MPath_Tools\Control\Packet as Control_Packet;

use \Nanoserv\Core as Nanoserv;

class Scoreboard {

	const STATUS_UP = 1;
	const STATUS_DOWN = 2;
	
	static private $instance;
	
	public $gateways = array();
	public $logs = array();
	public $updates = array();

	private function __construct() {

	}
	
	static public function Get($new = false) {

		if ($new || !isset(self::$instance)) {

			self::$instance = Nanoserv::New_Shared_Object(new self());

		}

		return self::$instance;
	
	}

	public function Add_Log($gwname, $status, $latency = NULL) {

		$this->logs[$gwname][] = array($status, $latency);

	}
	
	public function Get_Logs() {

		$ret = $this->logs;

		$this->logs = array();

		return $ret;
	
	}

	public function Queue_Updates($gwname, array $updates) {

		if (!isset($this->updates[$gwname])) {

			$this->updates[$gwname] = $updates;

		} else foreach ($updates as $key => $val) {
			
			$this->updates[$gwname][$key] = $val;

		}
	
	}

	public function Get_Pending_Updates() {

		$ret = $this->updates;

		$this->updates = array();

		return $ret;
	
	}

}

function mpd_show_help() {

	echo "usage: {$GLOBALS['argv'][0]} [options]\n";
	echo "\n";
	echo "    -c <filename>      configuration file\n";
	echo "    -f                 run in foreground and send output to console\n";
	echo "    -n                 disable logging\n";
	echo "    -v                 show version number and exit (-V for extended)\n";
	echo "    -t                 test configuration and exit\n";
	echo "    -d                 dump tables and rules to console and exit (for testing)\n";
	echo "    -h                 this help page\n";
	echo "\n";

}

function mpd_check_deps() {

	if (version_compare(phpversion(), "5.3") < 0) {

		throw new Exception("PHP version >= 5.3 needed");

	}
	
	$needed = array("pcntl", "posix", "sockets", "zlib");
	
	foreach ($needed as $ext) {

		if (!extension_loaded($ext)) {

			throw new Exception("php needs the '{$ext}' extension");

		}
	
	}
	
	return true;

}

function mpd_out($level, $s, Control_Server $except = NULL) {

	$level_str = Log::Level_To_String($level);
	
	if ($GLOBALS["foreground"]) {
	
		echo "{$level_str}: {$s}\n";

	}

	Control_Server::Send_Log($s, $level, $except);
	Log::Write($s, $level);

}

function mpd_debug($s, Control_Server $except = NULL) {

	return mpd_out(Log::LEVEL_DEBUG, $s, $except);
	
}

function mpd_notice($s, Control_Server $except = NULL) {

	return mpd_out(Log::LEVEL_NOTICE, $s, $except);

}

function mpd_warning($s, Control_Server $except = NULL) {

	return mpd_out(Log::LEVEL_WARNING, $s, $except);

}

function mpd_error($s, Control_Server $except = NULL) {

	return mpd_out(Log::LEVEL_ERROR, $s, $except);

}

function mpd_shutdown() {

	global $conf, $probe_pid;

	if (($conf->pidfile) && (file_exists($conf->pidfile)) && !unlink($conf->pidfile)) {

		mpd_warning("cannot delete pid file '{$conf->pidfile}'");
	
	}

	if ($probe_pid) {

		posix_kill($probe_pid, SIGTERM);
	
	}
	
	if ($conf->events && $conf->events->shutdown) {

		mpd_debug("executing shutdown events");
		mpd_exec($conf->events->shutdown->Get_Values());
	
	}
	
	mpd_notice("mpathd is shutting down\n");

}

function mpd_probe_thread() {

	pcntl_signal(SIGTERM, SIG_DFL);
	
	sleep(1);
	
	$gws = Gateway::Get_All(true);
	
	Probe::Initialize($gws);

	$sb = Scoreboard::Get();
	
	while (true) {
	
		$pstat = array();
		
		foreach (Probe::Run() as $gwname => $pdata) {

			list($status, $latency, $logs) = $pdata;

			$pstat[$gwname] = array($status, $latency);
		
			foreach ($logs as $probe_log) {

				$sb->Add_Log($gwname, $probe_log[0], $probe_log[1]);

			}
		
		}
		
		if ($pstat !== $old_pstat) {
		
			$sb->gateways = $old_pstat = $pstat;

		}

		foreach ($sb->Get_Pending_Updates() as $gwname => $updates) {

			$gw = Gateway::Get_By_Name($gwname);
			
			foreach ($updates as $key => $val) {

				try {
				
					switch ($key) {

						case "ip_addr":		$gw->ip_addr = $val;				break;
						case "network":		$gw->network = $val;				break;
						case "nexthop":		$gw->nexthop = $val;				break;
						case "interface":	$gw->interface = $val;				break;
						case "weight":		$gw->weight = $val;					break;
						case "state":		$gw->admin_state = $val;			break;
						case "latency":		$gw->max_latency = $val / 1000;		break;
						case "interval":	$gw->Get_Probe()->interval = $val;	break;
						case "timeout":		$gw->Get_Probe()->timeout = $val;	break;

					}
			
				} catch (Exception $e) {

					// Ignore errors
				
				}
			
			}

		}

		if (!Probe::Has_Waiting()) {
		
			Probe::Sleep_Until_Next_Probe();

		}

	}

}

function mpd_kill_handler() {

	exit();

}

function mpd_hup_handler() {

	mpd_reload();

}

function mpd_usr1_handler() {

	mpd_notice("received SIGUSR1, regenerating routing tables and rules");
	
	mpd_exec(mpd_regenerate_tables());

}

function mpd_regenerate_tables($source_only = false) {

	$ret = array();
	
	$ret[] = "ip rule flush";
	$ret[] = "ip rule add prio 10 table main";

	Iproute2_Rule::Set_First_Prio(11);

	if ($source_only) {

		Iproute2_Rule::Regenerate_Source();
		Rule::$default->Get_Iproute2_Rules();
	
	} else {
	
		Iproute2_Rule::Regenerate_All();

	}
	
	foreach (Iproute2_Table::Get_All() as $table) $ret = array_merge($ret, $table->Get_Commands());
	foreach (Iproute2_Rule::Get_All() as $rule) $ret = array_merge($ret, $rule->Get_Commands());

	$ret[] = "ip route flush cache";

	return $ret;

}

function mpd_exec($cmds, $bulk = true) {

	$cmds = (array)$cmds;
	
	if ($bulk) {

		\exec(implode(";", $cmds), $out, $retval);
	
		if ($retval !== 0) {

			mpd_debug("returned: " . trim(implode("\n", $out)));

		}
	
	} else {
	
		foreach ($cmds as $k => $cmd) {

			mpd_debug("exec: {$cmd}");

			$out = array();
			$retval = 0;
			
			\exec($cmd, $out, $retval);
			
			if ($retval !== 0) {

				mpd_debug("returned: " . trim(implode("\n", $out)));

			}
		
		}

	}

}

function mpd_delayed_init() {

	if (Gateway::All_Probed()) {
	
		mpd_exec(mpd_regenerate_tables());
		mpd_notice("routing tables and rules are set up");

	} else {

		Nanoserv::New_Timer(1, __FUNCTION__);

	}

}

// control functions

function mpd_show_gateways($search) {

	$ret = array();
	
	foreach (Gateway::Get_All(true) as $gw) {

		if (($search !== false) && (strpos($gw->name, $search) === false)) {

			continue;

		}
		
		$tmp = array();

		$tmp["name"] = $gw->name;
		$tmp["ip_addr"] = $gw->ip_addr;
		$tmp["network"] = $gw->network;
		$tmp["nexthop"] = $gw->nexthop;
		$tmp["interface"] = $gw->interface;
		$tmp["weight"] = $gw->weight;
		$tmp["state"] = $gw->Get_State_Str();
		$tmp["status"] = $gw->Get_Status_Str();

		try {
		
			$tmp["loss_pct_60sec"] = $gw->Get_Stats()->Get_Packet_Loss_Pct(60);

		} catch (Exception $e) {

			$tmp["loss_pct_60sec"] = false;

		}

		$ret[] = $tmp;
	
	}

	return $ret;

}

function mpd_show_gateway(array $names) {

	$ret = array();
	
	if (!$names) {

		foreach (Gateway::Get_All() as $gw) $names[] = $gw->name;
	
	}
	
	foreach ($names as $name) {

		if (!$gw = Gateway::Get_By_Name($name)) {

			return array("error" => "invalid gateway '{$name}'");

		}

		$tmp = array();

		$tmp["name"] = $gw->name;
		$tmp["ip_addr"] = $gw->ip_addr;
		$tmp["network"] = $gw->network;
		$tmp["nexthop"] = $gw->nexthop;
		$tmp["interface"] = $gw->interface;
		$tmp["weight"] = $gw->weight;

		$tmp["state"] = $gw->Get_State_Str();
		$tmp["status"] = $gw->Get_Status_Str();

		$tmp["disengage"] = $gw->disengage;
		$tmp["disable"] = $gw->disable;
		$tmp["activate"] = $gw->activate;
		$tmp["latency"] = $gw->max_latency * 1000;

		try {
		
			$probe = $gw->Get_Probe();
			
			$tmp["interval"] = $probe->interval;
			$tmp["timeout"] = $probe->timeout;

		} catch (Exception $e) {

			$tmp["interval"] = $tmp["timeout"] = false;
		
		}
		
		try {
		
			$stats = $gw->Get_Stats();
			
			$stats_dur = array(10, 30, 60, 300, 600, 1800, 3600);

			$tmp["stats"] = array();
			
			foreach ($stats_dur as $dur) {

				$tmp2 = array();

				foreach ($stats->Get_Probe_Data($dur) as $k => $v) {

					$tmp2[$k] = $v;

				}

				foreach ($stats->Get_Rtt_Data($dur) as $k => $v) {

					$tmp2[$k] = $v;

				}

				$tmp["stats"][$dur] = $tmp2;
			
			}
		
		} catch (Exception $e) {

			$tmp["stats"] = false;

		}
		
		$ret[] = $tmp;
	
	}

	return $ret;

}

function mpd_set_gateway($gwname, array $updates) {

	try {

		if (!$gw = Gateway::Get_By_Name($gwname)) {

			throw new Exception("no such gateway '{$gwname}'");

		}

		foreach ($updates as $key => $val) {

			try {
			
				switch ($key) {

					case "ip_addr":		$gw->ip_addr = $val;				break;
					case "network":		$gw->network = $val;				break;
					case "nexthop":		$gw->nexthop = $val;				break;
					case "interface":	$gw->interface = $val;				break;
					case "weight":		$gw->weight = $val;					break;
					case "state":		$gw->Set_Admin_State_Str($val);		break;
					case "latency":		$gw->max_latency = $val / 1000;		break;
					case "interval":	$gw->Get_Probe()->interval = $val;	break;
					case "timeout":		$gw->Get_Probe()->timeout = $val;	break;

					default:			throw new Exception("unknown parameter {$key}");
				
				}
		
			} catch (Exception $e) {

				throw new Exception("cannot set {$key}: {$e->getMessage()}");
			
			}
		
		}

		Scoreboard::Get()->Queue_Updates($gwname, $updates);
		
		return true;
	
	} catch (Exception $e) {

		return array("error" => $e->getMessage());
	
	}

}

function _mpd_rule_info(Rule $rule) {

	$ret = array();
	
	$ret["id"] = $rule->id;
	$ret["match"] = $rule->match;
	$ret["value"] = $rule->conf_value;
	$ret["current_value"] = $rule->value;
	$ret["default_policy"] = $rule->default_policy;
	$ret["expires"] = $rule->expires;
	$ret["routes"] = $rule->routes;
	$ret["table"] = $rule->table;
	
	return $ret;

}

function mpd_show_rule(array $rule = NULL) {

	$ret = array();
	
	try {
	
		if (!isset($rule)) {
			
			$ret[0] = _mpd_rule_info(Rule::$default);
		
			foreach (Rule::Get_All() as $rule) {

				$ret[$rule->id] = _mpd_rule_info($rule);

			}
		
		} else {
			
			$r = Rule::Get($rule["match"], $rule["value"]);
			
			$ret[$r->id] = _mpd_rule_info($r);

		}
	
	} catch (Exception $e) {

		$ret["error"] = $e->getMessage();
	
	}
	
	return $ret;

}

function mpd_add_rule(array $rule) {

	try {

		Rule::Create_From_Array($rule);
	
	} catch (Exception $e) {

		return array("error" => "cannot add rule: {$e->getMessage()}");
	
	}

	return true;

}

function mpd_insert_rule(array $rule, $position) {

	try {

		if ($position < 1) {

			throw new Exception("cannot insert rule at invalid position {$position}");

		}
		
		Rule::Create_From_Array($rule)->Reorder($position);
	
	} catch (Exception $e) {

		return array("error" => "cannot add rule: {$e->getMessage()}");
	
	}

	return true;

}

function mpd_del_rule(array $rule) {

	try {

		Rule::Get($rule["match"], $rule["value"])->Remove();

	} catch (Exception $e) {

		return array("error" => "cannot delete rule: {$e->getMessage()}");
	
	}

	return true;

}

function mpd_set_rule(array $rule) {

	try {

		$r = Rule::Get($rule["match"], $rule["value"]);

		foreach ($rule as $k => $v) {

			switch ($k) {

				case "match":			$r->Set_Match($v);				break;
				case "value":			$r->Set_Value($v);				break;
				case "via":				$r->Set_Routes($v);				break;
				case "default_policy":	$r->Set_Default_Policy($v);		break;
				case "expires":			$r->Set_Expires($v);			break;

				default:				throw new Exception("invalid rule param {$k}");
			
			}
		
		}
	
	} catch (Exception $e) {

		return array("error" => "cannot set rule: {$e->getMessage()}");
	
	}

	return true;

}

function mpd_get_log($filename) {

	try {

		$log = Log::Get_By_Filename($filename);

		if (!$log) {

			throw new Exception("unknown log file '{$filename}'");

		}
	
	} catch (Exception $e) {

		return array("error" => "cannot get log: {$e->getMessage()}");

	}

	$ret = array(	"filename"	=> $log->filename,
					"level"		=> $log->level);

	return $ret;

}

function mpd_set_log(array $log_arr) {

	try {

		$log = Log::Get_By_Filename($log_arr["filename"]);

		if (!$log) {

			throw new Exception("unknown log file '{$log_arr['filename']}'");

		}
	
		if (isset($log_arr["level"])) {

			$log->level = $log_arr["level"];

		}
	
	} catch (Exception $e) {

		return array("error" => "cannot get log: {$e->getMessage()}");

	}

	return true;

}

function mpd_reload() {

	global $conffile, $nologging, $conf, $sb, $probe_pid;
	
	try {

		mpd_notice("reloading config file '{$conffile}'");
		$conf = Conf::Parse($conffile);

		if (!$nologging) {
		
			mpd_debug("initializing logging facility");
			Log::Initialize($conf);

		}
		
		Gateway::Initialize($conf);
		Rule::Initialize($conf);

		Iproute2_Table_Info::Initialize();
		Iproute2_Realm::Initialize();
		Iproute2_Protocol::Initialize();

		Gateway::Prepare_All();
		Probe::Initialize();

		mpd_notice("restarting gateway probes");

		posix_kill($probe_pid, SIGTERM);
		
		mpd_exec(mpd_regenerate_tables(true));
		Nanoserv::New_Timer(1, __NAMESPACE__ . "\\mpd_delayed_init");

		$sb = Scoreboard::Get(true);
		$probe_pid = Nanoserv::Fork();

		if ($probe_pid === 0) {

			mpd_probe_thread();
			exit();

		} else if ($probe_pid === false) {

			mpd_error("cannot fork() probe process");
			exit(-4);

		}
	
		return true;
	
	} catch (Exception $e) {

		return array("error" => "cannot reload configuration: {$e->getMessage()}");
	
	}

}

function mpd_show_control() {

	$ret = array();
	
	$t = time();
	
	foreach (Control_Server::Get_All() as $ctl) {

		$ret[] = array("id" => $ctl->socket->id, "peer" => $ctl->peer, "agent" => $ctl->agent, "connect_t" => $ctl->connect_t, "idle_t" => $t - $ctl->last_activity_t);
	
	}

	return $ret;

}

function mpd_show_status() {

	global $pid, $probe_pid, $conffile, $started_t;
	
	$ret = array();

	$ret["version"] = VERSION;
	$ret["pid"] = $pid;
	$ret["probe_pid"] = $probe_pid;
	$ret["config_file"] = $conffile;
	$ret["started_t"] = $started_t;
	
	return $ret;

}

function mpd_handle_control_packet(Control_Server $ctl, Control_Packet $pkt) {

	try {

		$prefix = "[ctl:{$ctl->socket->id}] ";
		
		switch ($pkt->type) {

			case Control_Packet::TYPE_HELLO:
			mpd_debug("{$prefix}accepted control connection from '{$ctl->peer}' ({$pkt->message['agent']})", $ctl);
			break;
			
			case Control_Packet::TYPE_STOP:
			mpd_notice("{$prefix}received stop command" . ($pkt->message ? " ({$pkt->message})" : ""), $ctl);
			exit();
			break;

			case Control_Packet::TYPE_COMMIT:
			mpd_debug("{$prefix}received commit command, regenerating routing tables and rules", $ctl);
			$tmp = mpd_regenerate_tables();
			$ret = mpd_exec($tmp);
			$ctl->Send(Control_Packet::TYPE_COMMIT, array(count($tmp), count($ret)));
			break;
			
			case Control_Packet::TYPE_REFRESH:
			mpd_debug("{$prefix}received refresh command", $ctl);
			$GLOBALS["rules_refresh_t"] = 0;
			break;
			
			case Control_Packet::TYPE_GATEWAY_LIST:
			mpd_debug("{$prefix}received gateway list command", $ctl);
			$ctl->Send(Control_Packet::TYPE_GATEWAY_LIST, mpd_show_gateways($pkt->message));
			break;

			case Control_Packet::TYPE_GATEWAY_GET:
			mpd_debug("{$prefix}received gateway get command", $ctl);
			$ctl->Send(Control_Packet::TYPE_GATEWAY_GET, mpd_show_gateway($pkt->message));
			break;

			case Control_Packet::TYPE_GATEWAY_SET:
			mpd_debug("{$prefix}received gateway set command ({$pkt->message['gateway']})", $ctl);
			$ctl->Send(Control_Packet::TYPE_GATEWAY_SET, mpd_set_gateway($pkt->message["gateway"], $pkt->message["updates"]));
			break;
			
			case Control_Packet::TYPE_RULE_GET:
			mpd_debug("{$prefix}received rule get command", $ctl);
			$ctl->Send(Control_Packet::TYPE_RULE_GET, mpd_show_rule($pkt->message));
			break;
		
			case Control_Packet::TYPE_RULE_ADD:
			mpd_debug("{$prefix}received rule add command", $ctl);
			$ctl->Send(Control_Packet::TYPE_RULE_ADD, mpd_add_rule($pkt->message));
			break;
		
			case Control_Packet::TYPE_RULE_INSERT:
			mpd_debug("{$prefix}received rule insert command", $ctl);
			$ctl->Send(Control_Packet::TYPE_RULE_INSERT, mpd_insert_rule($pkt->message["rule"], $pkt->message["position"]));
			break;
			
			case Control_Packet::TYPE_RULE_DEL:
			mpd_debug("{$prefix}received rule del command", $ctl);
			$ctl->Send(Control_Packet::TYPE_RULE_DEL, mpd_del_rule($pkt->message));
			break;
		
			case Control_Packet::TYPE_RULE_SET:
			mpd_debug("{$prefix}received rule set command", $ctl);
			$ctl->Send(Control_Packet::TYPE_RULE_SET, mpd_set_rule($pkt->message));
			break;
		
			case Control_Packet::TYPE_LOG_GET:
			mpd_debug("{$prefix}received log get command", $ctl);
			$ctl->Send(Control_Packet::TYPE_LOG_GET, mpd_get_log($pkt->message));
			break;
		
			case Control_Packet::TYPE_LOG_SET:
			mpd_debug("{$prefix}received log set command", $ctl);
			$ctl->Send(Control_Packet::TYPE_LOG_SET, mpd_set_log($pkt->message));
			break;

			case Control_Packet::TYPE_RELOAD:
			mpd_debug("{$prefix}received reload command", $ctl);
			$ctl->Send(Control_Packet::TYPE_RELOAD, mpd_reload());
			break;
		
			case Control_Packet::TYPE_CONTROL_LIST:
			mpd_debug("{$prefix}received control list command", $ctl);
			$ctl->Send(Control_Packet::TYPE_CONTROL_LIST, mpd_show_control());
			break;
			
			case Control_Packet::TYPE_STATUS:
			mpd_debug("{$prefix}received status get command", $ctl);
			$ctl->Send(Control_Packet::TYPE_STATUS, mpd_show_status());
			break;
			
			default:
			$ctl->Send($pkt->type, array("error" => "control command not implemented"));
			break;
		
		}

	} catch (Exception $e) {

		mpd_warning("error while handling control packet: {$e->getMessage()}");
		$ctl->Disconnect();
	
	}

}

function mpd_handle_control() {

	foreach (Control_Server::Get_All() as $ctl) {

		while ($pkt = $ctl->Get_Packet()) {

			mpd_handle_control_packet($ctl, $pkt);
		
		}
	
	}

}

$opts = getopt("c:vVhfntd");

if (isset($opts["h"])) {

	mpd_show_help();
	exit();

}

if (isset($opts["v"]) || isset($opts["V"])) {

	if (isset($opts["V"])) {

		echo "PHP      " . phpversion() . "\n";
		echo "Nanoserv " . Nanoserv::VERSION . "\n";
		echo "mpathd   " . VERSION . "\n";
	
	} else {
	
		echo VERSION . "\n";

	}

	exit();

}

if (isset($opts["c"])) {

	$conffile = $opts["c"];

} else {

	$conffile = Conf::DEFAULT_PATH . DIRECTORY_SEPARATOR . DEFAULT_CONF_FILENAME;

}

$foreground = isset($opts["f"]) || isset($opts["t"]);
$nologging = isset($opts["n"]) || isset($opts["t"]) || isset($opts["d"]);

// Init

try {

	mpd_debug("loading config file '{$conffile}'");
	$conf = Conf::Parse($conffile);

	if (!$nologging) {
	
		mpd_debug("initializing logging facility");
		Log::Initialize($conf);

	}
	
	mpd_notice("mpathd/" . VERSION . " (c) 2008-2010 by sIX <six at aegis-corp.org>");

	mpd_check_deps();
	
	mpd_debug("initializing gateways"); 
	Gateway::Initialize($conf);

	mpd_debug("loading rules");
	Rule::Initialize($conf);
	
	mpd_debug("loading iproute2 data");
	Iproute2_Table_Info::Initialize();
	Iproute2_Realm::Initialize();
	Iproute2_Protocol::Initialize();

	mpd_debug("preparing gateways");
	Gateway::Prepare_All();
	Probe::Initialize();

} catch (Tokenizer_Exception $e) {

	mpd_error("config syntax error: {$e->getMessage()}");
	exit(-3);
	
} catch (Exception $e) {

	mpd_error("error: {$e->getMessage()}");
	exit(-3);

}

if (isset($opts["t"])) {

	mpd_notice("configuration check ok");
	exit(0);

}

if (isset($opts["d"])) {

	foreach (mpd_regenerate_tables() as $cmd) {

		echo $cmd . "\n";

	}
	
	exit(0);

}

// Detach from tty

if (!$foreground) {

	error_reporting(0);

	$pid = pcntl_fork();

	if ($pid < 0) {

		mpd_error("cannot fork()");
		exit(-4);

	} else if ($pid > 0) {

		exit();

	}

	fclose(STDIN);
	fclose(STDOUT);
	fclose(STDERR);

	$stdin = fopen("/dev/null", "r");
	$stdout = fopen("/dev/null", "w");
	$stderr = fopen("/dev/null", "w");
	
	posix_setsid();

}

mpd_notice("setting up source routing tables");
mpd_exec(mpd_regenerate_tables(true));

// Initialize probe thread

mpd_notice("starting gateway probes");

$sb = Scoreboard::Get(true);
$probe_pid = Nanoserv::Fork();

if ($probe_pid === 0) {

	mpd_probe_thread();
	exit();

} else if ($probe_pid === false) {

	mpd_error("cannot fork() probe process");
	exit(-4);

}

mpd_debug("probe process has been created (pid={$probe_pid})");

if ($conf->events && $conf->events->startup) {

	mpd_debug("executing startup events");
	mpd_exec($conf->events->startup->Get_Values());

}

register_shutdown_function(__NAMESPACE__ . "\\mpd_shutdown");
pcntl_signal(SIGTERM, __NAMESPACE__ . "\\mpd_kill_handler");
if ($foreground) pcntl_signal(SIGINT, __NAMESPACE__ . "\\mpd_kill_handler");
pcntl_signal(SIGHUP, __NAMESPACE__ . "\\mpd_hup_handler");
pcntl_signal(SIGUSR1, __NAMESPACE__ . "\\mpd_usr1_handler");

if ($conf->control && !$conf->control->disabled) {

	if ($tmp = $conf->control->listen) {

		$addrs = $tmp->Get_Values();

	} else {

		$addrs = array(Control_Server::DEFAULT_LISTEN_ADDR);

	}

	foreach ($addrs as $addr) {
	
		mpd_notice("starting control listener at '{$addr}'");

		try {
		
			Nanoserv::New_Listener("tcp://{$addr}", "\\MPath_Tools\\Control\\Server_Connection")->Activate();

		} catch (Exception $e) {

			mpd_warning("cannot start control server: {$e->getMessage()}");

		}

	}

}

$pid = posix_getpid();

if (($conf->pidfile) && (!file_put_contents($conf->pidfile, "{$pid}\n"))) {

	mpd_error("cannot write pid file '{$conf->pidfile}'");
	exit(-5);

}

Nanoserv::New_Timer(1, __NAMESPACE__ . "\\mpd_delayed_init");

$started_t = time();

// Main loop

while (true) {

	$state_changed = false;
	
	pcntl_signal_dispatch();
	
	Nanoserv::Run(1);

	mpd_handle_control();

	$t = time();
	
	foreach ($sb->gateways as $name => $sbdata) {

		list($status, $latency, $probes_log) = $sbdata;
		
		$gw = Gateway::Get_By_Name($name);

		if ($gw->status !== $gw->Get_Status_From_Probe($status, $latency)) {

			$tmp = $gw->status;
			
			$gw->Set_Status($status, $latency);

			if ($tmp === Gateway::STATUS_UNKNOWN) {

				mpd_notice("gateway '{$gw->name}' is {$gw->Get_State_Str()} [{$gw->Get_Status_Str()}]");			
			
			} else {
			
				mpd_debug("gateway '{$gw->name}' probe {$gw->Get_Status_Str()}");

			}
		
		} else {

			$gw->Set_Latency($latency);

		}
		
		if ($gw->Update_State($t)) {

			mpd_notice("gateway '{$gw->name}' is now {$gw->Get_State_Str()} [{$gw->Get_Status_Str()}]");

			if ($gw->events) {
			
				switch ($gw->Get_State()) {

					case Gateway::STATE_ACTIVE:			$ev_type = Gateway::EVENT_ACTIVATE;		break;
					case Gateway::STATE_DISENGAGED:		$ev_type = Gateway::EVENT_DISENGAGE;	break;
					case Gateway::STATE_DISABLED:		$ev_type = Gateway::EVENT_DISABLE;		break;

					default:							$ev_type = false;
				
				}
				
				if ($evs = $gw->events[$ev_type]) {

					$arg_gw = escapeshellarg($gw->name);
					$arg_state_num = escapeshellarg($gw->Get_State());
					$arg_state_str = escapeshellarg($gw->Get_State_Str());
					$arg_status_str = escapeshellarg($gw->Get_Status_Str());

					$args = "{$arg_gw} {$arg_state_num} {$arg_state_str} {$arg_status_str}";

					foreach ($evs as $ev) {
					
						mpd_exec("{$ev} {$args}");

					}
				
				}
			
			}
			
			$state_changed = true;
		
		}
	
		if (($rules_refresh_t + 3600) < $t) {

			if (Rule::Refresh_All()) {

				mpd_debug("refreshed rules, some resolved hostnames have changed");

				$state_changed = true;

			}
			
			$rules_refresh_t = $t;
		
		}
	
	}

	foreach ($sb->Get_Logs() as $name => $probe_logs) {

		foreach ($probe_logs as $probe_log) {

			list($status, $latency) = $probe_log;

			Stats::Add_Gateway_Data($name, $status, $latency);
		
		}
	
	}
	
	if ($expired = Rule::Do_Maintenance($t)) {

		mpd_debug("{$expired} expired rule(s) removed");
		
		$state_changed = true;
	
	}
	
	if ($state_changed) {
	
		mpd_exec(mpd_regenerate_tables());

	}

	// Watch for the probe process

	if (pcntl_waitpid($probe_pid, $tmp, WNOHANG) !== 0) {

		mpd_notice("probe process has gone away, trying to respawn");
		
		$probe_pid = Nanoserv::Fork();

		if ($probe_pid === 0) {

			mpd_probe_thread();
			exit();

		} else if ($probe_pid === false) {

			mpd_error("cannot fork() probe process");
			exit(-4);

		} else {

			mpd_notice("probe restarted (pid={$probe_pid})");

		}
	
	}

}

?>
