#!/usr/local/bin/php
<?php

/**
 *
 * mpath-tools - failover and load balancing between multiple ISP made easy
 * 
 * Copyright (C) 2008-2010 Vincent Negrier aka. sIX <six at aegis-corp.org>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 *
 * @package mpath-tools
 * @subpackage mping
 */

namespace MPath_Tools;

require "mpath-tools/inc/config.php";
require "mpath-tools/inc/getopt.php";
require "mpath-tools/inc/gateway.php";
require "mpath-tools/inc/ping.php";

use \Exception;
use \MPath_Tools\Config\Core as Conf;

const DEFAULT_CONF_FILENAME = "mping.conf";

const DEFAULT_COUNT = NULL;
const DEFAULT_INTERVAL = 1.0;
const DEFAULT_SIZE = 56;
const DEFAULT_PATTERN = "\x55\xAA";
const DEFAULT_TIMEOUT = 2.0;
const DEFAULT_DEADLINE = NULL;

error_reporting(0);

function mp_show_help() {

	echo "usage: {$GLOBALS['argv'][0]} [options] destination\n";
	echo "\n";
	echo "    -c <count>        send <count> requests and exit\n";
	echo "    -i <interval>     wait <interval> seconds between requests\n";
	echo "    -q                quiet mode\n";
	echo "    -s <size>         send <size> data bytes (default is 56)\n";
	echo "    -p <pattern>      repeat <pattern> up to data size\n";
	echo "    -W <timeout>      wait <timeout> seconds for a response\n";
	echo "    -w <timeout>      exit after <timeout> seconds\n";
	echo "    -4                force ipv4\n";
	echo "    -6                force ipv6\n";
	echo "    -g <gateways>     specify a list of comma separated gw names\n";
	echo "    -C                use ansi colors\n";
	echo "    -m <filename>     configuration file\n";
	echo "    -h                this help page\n";
	echo "\n";

}

function mp_get_gw_idx_by_name($name) {

	global $use_gws;

	if ($name === "local") {

		return false;

	}
	
	foreach ($use_gws as $gw) {

		if ($gw->name === $name) return $gw->idx;

	}

	return false;

}

function mp_ms_format($t) {

	$ms = $t * 1000;

	if ($ms >= 100) {

		return number_format($ms, 0);

	} else if ($ms >= 10) {

		return number_format($ms, 1);

	} else if ($ms >= 1) {

		return number_format($ms, 2);

	} else {

		return number_format($ms, 3);

	}

}

function mp_on_reply($ok, $addr, $source, $time, $reply) {

	if (!$GLOBALS["quiet"]) {
		
		if (!$GLOBALS["is_local"]) {
		
			$s = "[" . str_pad($source, 6, " ") . "] ";

		} else {

			$s = "";

		}
		
		if ($ok) {
		
			if (($GLOBALS["color"]) && (($idx = mp_get_gw_idx_by_name($source)) !== false)) {

				$s .= mp_get_color("gateways", mp_get_gw_idx_by_name($source));

			}
			
			$s .= strlen($reply->data) . " bytes from {$addr}: icmp_seq={$reply->seq} ttl={$reply->ttl} time=" . mp_ms_format($time) . " ms";

		} else {

			if ($GLOBALS["color"]) {

				$s .= mp_get_color("alert");
			
			}
			
			if ($addr) {
			
				$s .= "from {$addr}: ";
				
			}
			
			$s .= $time;
		
		}

		if ($GLOBALS["color"]) {

			$s .= mp_get_color("reset");

		}
		
		$s .= "\n";
		
		echo $s;
	
	}

	pcntl_signal_dispatch();

}

function mp_display_stats() {

	global $stats;
	
	if ($GLOBALS["color"]) {

		echo mp_get_color("reset");
	
	}
	
	echo "--- {$GLOBALS['host_name']} " . ($GLOBALS["gw_count"] > 1 ? "multipath " : "") . "ping statistics ---\n";

	$gws = array();
	
	foreach ($GLOBALS["use_gws"] as $gw) {

		$gws[$gw->name] = array("rcvd" => 0, "tot" => 0);

	}
	
	$xmit = count($stats);

	$ret = false;
	
	foreach ($stats as $replies) {

		foreach ($replies as $gw => $time) {

			$ret = true;
			
			$gws[$gw]["rcvd"]++;
			$gws[$gw]["tot"] += $time;

			if ((!isset($gws[$gw]["rtt_min"])) || ($time < $gws[$gw]["rtt_min"])) {

				$gws[$gw]["rtt_min"] = $time;

			}
		
			if ((!isset($gws[$gw]["rtt_max"])) || ($time > $gws[$gw]["rtt_max"])) {

				$gws[$gw]["rtt_max"] = $time;

			}
		
		}
	
	}
	
	foreach ($gws as $gwname => $gwstats) {

		if ($gwstats["rcvd"] > 1) {
		
			$gws[$gwname]["rtt_avg"] = $gwstats["tot"] / $gwstats["rcvd"];

			$dev = 0;
			
			foreach ($stats as $replies) {

				if (isset($replies[$gwname])) {

					$dev += pow($replies[$gwname] - $gws[$gwname]["rtt_avg"], 2);
				
				}
			
			}

			$dev /= $gwstats["rcvd"] - 1;

		} else {

			$dev = 0;

		}
			
		$gws[$gwname]["dev"] = sqrt($dev);
	
	}
	
	foreach ($gws as $gwname => $gwstats) {

		if ($xmit) {

			$loss = 100 - ($gwstats["rcvd"] / $xmit * 100);

		} else {

			$loss = 0;

		}
		
		if (!$GLOBALS["is_local"]) {
		
			echo str_pad($gwname, 8, " ") . " ";

		}
	
		echo "{$xmit} packets transmitted, {$gwstats['rcvd']} received, ";
		echo number_format($loss) . "% packet loss\n";

		if ($gwstats["rcvd"]) {
		
			if (!$GLOBALS["is_local"]) {
			
				echo str_pad("", 8, " ") . " ";
		
			}
			
			echo "rtt min/avg/max/mdev = " . number_format($gwstats["rtt_min"] * 1000, 3) . "/" . number_format($gwstats["rtt_avg"] * 1000, 3) . "/" . number_format($gwstats["rtt_max"] * 1000, 3) . "/" . number_format($gwstats["dev"] * 1000, 3) . " ms\n";

		}
	
	}
	
	return $ret;

}

function mp_check_deps() {

	if (version_compare(phpversion(), "5.3") < 0) {

		throw new Exception("PHP version >= 5.3 needed");

	}
	
	$needed = array("sockets", "pcntl");
	
	foreach ($needed as $ext) {

		if (!extension_loaded($ext)) {

			throw new Exception("php needs the '{$ext}' extension");

		}
	
	}
	
	return true;

}

function mp_on_break() {

	echo "\n";
	mp_display_stats();

}

function mp_signal_handler($sig) {

	if ($sig === SIGINT) {

		mp_on_break();
		die();
	
	}

}

function mp_on_shutdown() {

	if ($GLOBALS["color"]) {

		echo mp_get_color("reset");
	
	}

}

function mp_get_color($name, $idx=false) {

	global $conf;
	
	$code = chr(27) . "[";

	if ($tmp = $conf->colors->$name) {

		if ($idx === false) {

			$conf_clr = $tmp->Get_Value();

		} else {

			$c = count($tmp->Get_Values());
			
			$conf_clr = $tmp->Get_Value($idx % $c);

		}
	
	}

	if (isset($conf_clr)) {

		$code .= $conf_clr;
	
	} else {
	
		switch ($name) {

			case "reset":		$code .= "0m";						break;
			case "alert":		$code .= "0;31m";					break;
			case "altern":		$code .= $idx % 2 ? "0m" : "1;30m";	break;

		}

	}
	
	return $code;

}

function mp_is_local($addr) {

	global $conf;

	$a = new Address($addr);
	
	if ($a->family === AF_INET6) {

		// This is as hack and has to be fixed some time
		
		return true;

	} else if ($a->Is_Local()) {

		return true;

	} else if (!$conf->local) {

		return false;

	}

	foreach ($conf->local->Get_Values() as $network) {

		if ($a->Match_Network($network)) {

			return true;

		}

	}

	return false;

}

$opts = getopt("c:i:qs:p:w:W:g:m:46Ch-?");

if (isset($opts["?"])) {

	if (is_array($opts["?"])) {

		$param = $opts["?"][0];

	} else {

		$param = $opts["?"];

	}
	
	echo "error: invalid parameter '{$param}', try -h for help\n";
	exit(-1);

}

if (!isset($opts["-"]) || isset($opts["h"])) {

	mp_show_help();
	exit();

}

$conffile = $opts["m"] ? $opts["m"] : (Conf::DEFAULT_PATH . DIRECTORY_SEPARATOR . DEFAULT_CONF_FILENAME);

try {

	mp_check_deps();
	
	$conf = Conf::Parse($conffile);

	Gateway::Initialize($conf, true);

	$def_interval = $conf->defaults->interval ? $conf->defaults->interval[0] : DEFAULT_INTERVAL;
	$def_size = $conf->defaults->size ? $conf->defaults->size[0] : DEFAULT_SIZE;
	$def_pattern = $conf->defaults->pattern ? $conf->defaults->pattern[0] : DEFAULT_PATTERN;
	$def_timeout = $conf->defaults->timeout ? $conf->defaults->timeout[0] : DEFAULT_TIMEOUT;
	$def_color = $conf->defaults->color ? true : false;
	
	$count = (int)$opts["c"] ? (int)$opts["c"] : DEFAULT_COUNT;
	$interval = $opts["i"] ? $opts["i"] : $def_interval;
	$quiet = isset($opts["q"]);
	$size = (int)$opts["s"] ? (int)$opts["s"] : $def_size;
	$pattern = isset($opts["p"]) ? $opts["p"] : $def_pattern;
	$deadline = $opts["w"] ? $opts["w"] : DEFAULT_DEADLINE;
	$timeout = $opts["W"] ? $opts["W"] : $def_timeout;
	$gw_list = $opts["g"] ? explode(",", $opts["g"]) : NULL;
	$color = isset($opts["C"]) ? true : $def_color;


} catch (\Tokenizer\Exception $e) {
	
	echo "config syntax error: {$e->getMessage()}\n";
	exit(-1);

} catch (Exception $e) {

	echo "error: {$e->getMessage()}\n";
	exit(-1);

}

$host = $opts["-"];

if (is_array($host)) {

	echo "cannot ping multiple hosts\n";
	exit(-2);

}

if (Address::Is_IP($host)) {

	$host_ip = $host;
	$host_name = gethostbyaddr($host_ip);

} else {

	$host_name = $host;

	if (isset($opts["4"])) {
		
		$rec_type = DNS_A;

	} else if (isset($opts["6"])) {
		
		$rec_type = DNS_AAAA;

	} else {

		$rec_type = DNS_A | DNS_AAAA;
	
	}
	
	$host_dnsinfo = dns_get_record($host_name, $rec_type);

	foreach ($host_dnsinfo as $tmp) {

		switch ($tmp["type"]) {

			case "A":		$host_ip4 = $tmp["ip"];		break;
			case "AAAA":	$host_ip6 = $tmp["ipv6"];	break;

		}
		
		$host_ip = $host_ip6
		or $host_ip = $host_ip4;

	}

	if (!isset($host_ip) && !isset($opts["4"]) && !isset($opts["6"])) $host_ip = gethostbyname($host_name);
	
	if (!Address::Is_IP($host_ip)) {

		echo "unknown host {$host}\n";
		exit(-2);
	
	}

	$tmp = gethostbyaddr($host_ip);

	if ($tmp !== $host_ip) {

		$host_name = $tmp;

	}

}

$host_addr = new Address($host_ip);

if ($is_local = mp_is_local($host_ip)) {

	switch ($host_addr->family) {

		case AF_INET:	$use_gws = array(new Gateway("local", "0.0.0.0"));	break;
		case AF_INET6:	$use_gws = array(new Gateway("local", "::"));
	
	}

} else {

	$use_gws = array();	
	
	if (isset($gw_list)) {

		foreach ($gw_list as $gwname) {

			$use_gws[] = Gateway::Get_By_Name(trim($gwname));
		
		}

	} else if ($conf->defaults->gateways) {
		
		foreach ($conf->defaults->gateways as $gwname) {

			$use_gws[] = Gateway::Get_By_Name($gwname);

		}

	} else {

		foreach (Gateway::Get_All() as $gw) {

			if (($gw->admin_state !== Gateway::STATE_DISENGAGED) && ($gw->admin_state !== Gateway::STATE_DISABLED)) {

				$use_gws[] = $gw;

			}
		
		}

	}

}

$gw_addrs = array();

foreach ($use_gws as $gw) {

	$gw_addrs[$gw->name] = $gw->ip_addr;

}

$gw_count = count($use_gws);

echo "MPING {$host_name} ({$host_ip}) {$size} bytes of data" . ($gw_count > 1 ? " using {$gw_count} paths" : "") . ".\n";

try {

	$p = new Ping();
	$p->Set_Socket_Family($host_addr->family);
	
	register_shutdown_function(__NAMESPACE__ . "\\mp_on_shutdown");

	pcntl_signal(SIGINT, __NAMESPACE__ . "\\mp_signal_handler");

	$stats = array();

	$i = 0;

	if (isset($deadline)) {

		$exit_mt = microtime(true) + $deadline;

	}

	do {
		
		$mt = microtime(true);
		
		$stats[] = $p->Do_Ping($host_ip, $gw_addrs, $size, $pattern, $timeout, __NAMESPACE__ . "\\mp_on_reply");

		if (((isset($count)) && (++$i >= $count)) || (isset($exit_mt) && ($mt >= $exit_mt))) {

			break;

		}
		
		$wait = $interval - (microtime(true) - $mt);

		if ($wait > 0) {

			usleep($wait * 1000000);

		}
		
		if (count($stats) > 10000) {

			array_shift($stats);

		}

		pcntl_signal_dispatch();
	
	} while (true);

	echo "\n";

	exit(mp_display_stats() ? 0 : 1);

} catch (Exception $e) {

	echo "error: {$e->getMessage()}\n";
	exit(-2);

}

?>
